#!/bin/bash

# Source the profile script which should contain at least a JAVA_HOME
if [ -f ~/.nsp_perf_profile ]; then
	. ~/.nsp_perf_profile
fi

# Ensure JAVA_HOME is set
if [ -z "$JAVA_HOME" ]; then
	echo "Error: JAVA_HOME environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export JAVA_HOME

# Ensure MAVEN_HOME is set
if [ -z "$MAVEN_HOME" ]; then
	echo "Error: MAVEN_HOME environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export MAVEN_HOME

# Ensure M3 is set
if [ -z "$M3" ]; then
	echo "Error: M3 environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export M3

# Ensure SVN_HOME is set
if [ -z "$SVN_HOME" ]; then
	echo "Error: SVN_HOME environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export SVN_HOME

SCRIPT_NAME=`basename $0`

# for checking errors - and exiting if any - call as
#  check_errs $? 'Some error message'
check_errs()
{
  # Function. Parameter 1 is the return code
  # Para. 2 is text to display on failure.
  if [ "$1" -ne "0" ]; then
	echo "ERROR # ${1} : ${2}"
	exit ${1}
  fi
}

# Print usage
usage()
{
  # Function.
	echo "usage: ${SCRIPT_NAME} [-h <hosts configuration file> -p <Port number>] <NSP Component> <Testplan> <Distribution> "
}

# Read optional arguments using getopts
NODES=1
THREADS=20
HOSTS=hosts.properties
PORT=9012
while getopts “h:p:t:n:” OPTION; do
	case $OPTION in
		h)
			HOSTS=$OPTARG
			;;
		p)
			PORT=$OPTARG
			;;
		t)
			THREADS=$OPTARG
			;;
		n)
			NODES=$OPTARG
			;;
		?)
			usage
			exit
			;;
	esac
done

# Read mandatory arguments using shift and $1
shift $(($OPTIND - 1))
COMPONENT=$1
shift
TESTPLAN=$1
shift
DISTRIBUTION=$1

# Ensure mandatory arguments are provided
if [[ -z $COMPONENT ]] || [[ -z $DISTRIBUTION ]] || [[ -z $TESTPLAN ]]; then
	usage
	exit 2
fi

# Identification variables
MASTER=`hostname`
STARTED=`date +%Y%m%d_%H%M%S`

# Build file names
TESTPLAN_NAME=${COMPONENT}_${TESTPLAN}_${DISTRIBUTION}_${MASTER}_${STARTED}.assembled.jmx
REPORT_NAME=${COMPONENT}_${TESTPLAN}_${DISTRIBUTION}_${MASTER}_master_${STARTED}

# A hassle to get current working dir
unamestr=`uname -s`
case "${unamestr}" in
	*Linux*)
		CWD=$(dirname $(readlink -f $0))
	;;
	*Darwin*)
		# readlink works different on OS X so we use a python script instead
		REALPATH=$(python -c 'import os,sys;print os.path.realpath(sys.argv[1])' $0)
		CWD=$(dirname $REALPATH)
	;;
	*CYGWIN*)
		# translate to full cygpath
		CWD=$(cygpath -m $(dirname $(readlink -f $(cygpath $0))))
	;;
	*)
		echo "Unable to detect OS from '${unamestr}'"
		exit 1
	;;
esac
pushd ${CWD} > /dev/null

# Verify input
if [[ ! -d ${CWD}/tests/${COMPONENT} ]]; then
	check_errs 3 "Component '${COMPONENT}' not found."
fi
if [[ ! -r ${CWD}/tests/${COMPONENT}/src/test/jmeter/templates/testplans/${TESTPLAN}.template.jmx ]]; then
	check_errs 3 "Testplan '${TESTPLAN}' not found."
fi
if [[ ! -r ${CWD}/tests/${COMPONENT}/src/test/jmeter/templates/distributions/${DISTRIBUTION}.template.jmx ]]; then
	check_errs 3 "Distribution '${DISTRIBUTION}' not found."
fi

echo "Preparing performance test of NSP component '${COMPONENT}'.."

find ${CWD} -name "*.assembled.jmx" | xargs rm -f
check_errs $? "Unable to prepare test plan directories"

find ${CWD} -name "*.jtl" | xargs rm -f
check_errs $? "Unable to prepare test result directories"

${M3}/mvn clean install > ${CWD}/mvn.clean.install.log
check_errs $? "Unable to prepare NSP JMeter Extensions, is Maven available?"

echo "Synchronizing execution with ${NODES} slave nodes:"
# Verify slave configuration
for ((i=1; i<=${NODES}; i++)); do
	SLAVE_NODE=`grep slave${i}.host ${HOSTS} | cut -d'=' -f2`
	SLAVE_PATH=`grep slave${i}.path ${HOSTS} | cut -d'=' -f2`
	if [[ -z "${SLAVE_NODE}" ]] || [[ -z "${SLAVE_PATH}" ]]; then
		check_errs 1 "Slave ${i} was not configured in ${HOSTS}"
	fi
	echo -e "\t${i}\t${SLAVE_NODE}:${SLAVE_PATH}"
done

# Copy the configuration to the slaves
for ((i=1; i<=${NODES}; i++)); do
	SLAVE_NODE=`grep slave${i}.host ${HOSTS} | cut -d'=' -f2`
	SLAVE_PATH=`grep slave${i}.path ${HOSTS} | cut -d'=' -f2`
	if [[ -n "${SLAVE_NODE}" ]] && [[ -n "${SLAVE_PATH}" ]]; then
		SLAVE_HOSTS=${COMPONENT}_${TESTPLAN}_${DISTRIBUTION}_${MASTER}_slave${i}_${STARTED}.hosts
		scp ${HOSTS} ${SLAVE_NODE}:${SLAVE_PATH}/${SLAVE_HOSTS} > /dev/null
		check_errs $? "Unable to distribute the hosts configuration to ${SLAVE_NODE}:${SLAVE_PATH}"
	fi
done

# Start all slaves
PREVIOUS_SLAVE_NODE=
ACTIVE_NODES=`expr ${NODES} + 1`
for ((i=1; i<=${NODES}; i++)); do
	SLAVE_NODE=`grep slave${i}.host ${HOSTS} | cut -d'=' -f2`
	SLAVE_PATH=`grep slave${i}.path ${HOSTS} | cut -d'=' -f2`
	if [[ -n "${SLAVE_NODE}" ]] && [[ -n "${SLAVE_PATH}" ]]; then
		if [[ "${PREVIOUS_SLAVE_NODE}" == "${SLAVE_NODE}" ]]; then
			echo "Delaying startup of multiple slaves on the same node: ${SLAVE_NODE}"
			sleep 20
		fi
		PREVIOUS_SLAVE_NODE="${SLAVE_NODE}"
		SLAVE_HOSTS=${COMPONENT}_${TESTPLAN}_${DISTRIBUTION}_${MASTER}_slave${i}_${STARTED}.hosts
		SLAVE_LOG=${COMPONENT}_${TESTPLAN}_${DISTRIBUTION}_${MASTER}_slave${i}_${STARTED}.log
		ssh ${SLAVE_NODE} "nohup ${SLAVE_PATH}/slave.sh ${MASTER} ${PORT} ${i} ${ACTIVE_NODES} ${COMPONENT} ${TESTPLAN} ${DISTRIBUTION} ${THREADS} ${STARTED} ${SLAVE_HOSTS} >${SLAVE_PATH}/${SLAVE_LOG} 2>&1 &" > /dev/null
		check_errs $? "Unable to start slave at ${SLAVE_NODE}:${SLAVE_PATH}"
	fi
done

JMETER_COMMON=`find ${CWD}/modules/jmeter-common -name "jmeter-common*.jar"`

# Build testplan
${JAVA_HOME}/bin/java -cp ${JMETER_COMMON} com.arosii.jmeter.utils.nsp.PrepareTestPlan ${CWD}/tests/${COMPONENT}/src/test/jmeter/templates/testplans/${TESTPLAN}.template.jmx ${CWD}/tests/${COMPONENT}/src/test/jmeter/templates/distributions/${DISTRIBUTION}.template.jmx ${CWD}/tests/${COMPONENT}/src/test/jmeter/${TESTPLAN_NAME} ${THREADS} ${ACTIVE_NODES} 0 ${HOSTS} > ${CWD}/java.PrepareTestPlan.log 2>&1
check_errs $? "Unable to create test plan with selected distribution"

# Synchronize with slaves
echo "Waiting for slave node(s)"
${JAVA_HOME}/bin/java -cp ${JMETER_COMMON} com.arosii.jmeter.utils.nsp.sync.SyncMaster ${NODES} 10 ${PORT} > ${CWD}/java.SyncMaster.log 2>&1
check_errs $? "Unable to synchronize execution with slaves"

NSP_HOST=`grep nsp.host ${HOSTS} | cut -d'=' -f2`
NSP_PORT=`grep nsp.port ${HOSTS} | cut -d'=' -f2`

# Start background process if present
if [ -f ${CWD}/tests/${COMPONENT}/start.sh ]; then
	echo "Starting background process.."
	pushd ${CWD}/tests/${COMPONENT} > /dev/null
	${CWD}/tests/${COMPONENT}/start.sh ${CWD} ${COMPONENT} ${TESTPLAN} ${DISTRIBUTION} ${NSP_HOST} ${NSP_PORT} > ${CWD}/background.start.log 2>&1 &
	popd > /dev/null
fi

pushd ${CWD}/tests/${COMPONENT} > /dev/null

# Run actual test plan
echo "Running performance test of NSP component '${COMPONENT}', this may take several minutes.."
echo "Using testplan '${TESTPLAN}' and distribution '${DISTRIBUTION}'"
echo "Started at `date +%Y-%m-%d_%H-%M-%S`"
${M3}/mvn verify > ${CWD}/mvn.verify.log
check_errs $? "Unable to run testplan"
echo "Finished at `date +%Y-%m-%d_%H-%M-%S`"

popd > /dev/null

# Synchronize with slaves
echo "Waiting for slave node(s) to finish"
${JAVA_HOME}/bin/java -cp ${JMETER_COMMON} com.arosii.jmeter.utils.nsp.sync.SyncMaster ${NODES} 120 ${PORT} >> ${CWD}/java.SyncMaster.log 2>&1
check_errs $? "Unable to synchronize report collection with slaves"

# Stop background process if present
if [ -f ${CWD}/tests/${COMPONENT}/stop.sh ]; then
	echo "Stopping background process.."
	pushd ${CWD}/tests/${COMPONENT} > /dev/null
	stop.sh ${TESTPLAN} ${NSP_HOST} ${NSP_PORT} > ${CWD}/background.stop.log 2>&1
	popd > /dev/null
fi

echo "Creating performance report.."

mkdir ${CWD}/${REPORT_NAME}
check_errs $? "Unable to create report directory"
mkdir ${CWD}/${REPORT_NAME}/master
check_errs $? "Unable to create report directory for master"

# Copy the reports from the slaves
for ((i=1; i<=${NODES}; i++)); do
	SLAVE_NODE=`grep slave${i}.host ${HOSTS} | cut -d'=' -f2`
	SLAVE_PATH=`grep slave${i}.path ${HOSTS} | cut -d'=' -f2`
	if [[ -n "${SLAVE_NODE}" ]] && [[ -n "${SLAVE_PATH}" ]]; then
		SLAVE_REPORT_NAME=${COMPONENT}_${TESTPLAN}_${DISTRIBUTION}_${MASTER}_slave${i}_${STARTED}.tar.gz
		SLAVE_LOG=${COMPONENT}_${TESTPLAN}_${DISTRIBUTION}_${MASTER}_slave${i}_${STARTED}.log
		mkdir ${CWD}/${REPORT_NAME}/slave${i} > /dev/null
		check_errs $? "Unable to create report directory for slave ${i}"
		scp ${SLAVE_NODE}:${SLAVE_PATH}/${SLAVE_REPORT_NAME} ${CWD}/${REPORT_NAME}/slave${i}/${SLAVE_REPORT_NAME} > /dev/null
		check_errs $? "Unable to collect performance report ${SLAVE_NODE}:${SLAVE_PATH}/${SLAVE_REPORT_NAME}"
		ssh ${SLAVE_NODE} "rm -f ${SLAVE_PATH}/${SLAVE_REPORT_NAME}" > /dev/null
		check_errs $? "Unable to remove performance report ${SLAVE_PATH}/${SLAVE_REPORT_NAME} on ${SLAVE_NODE}"
		scp ${SLAVE_NODE}:${SLAVE_PATH}/${SLAVE_LOG} ${CWD}/${REPORT_NAME}/slave${i}/${SLAVE_LOG} > /dev/null
		check_errs $? "Unable to collect slave log file ${SLAVE_NODE}:${SLAVE_PATH}/${SLAVE_LOG}"
		ssh ${SLAVE_NODE} "rm -f ${SLAVE_PATH}/${SLAVE_LOG}" > /dev/null
		check_errs $? "Unable to remove slave log file ${SLAVE_PATH}/${SLAVE_LOG} on ${SLAVE_NODE}"
		pushd ${CWD}/${REPORT_NAME}/slave${i} > /dev/null
		tar -xzf ${SLAVE_REPORT_NAME} > /dev/null
		check_errs $? "Unable to extract performance report ${CWD}/${REPORT_NAME}/slave${i}/${SLAVE_REPORT_NAME}"
		popd > /dev/null
		rm -f ${CWD}/${REPORT_NAME}/slave${i}/${SLAVE_REPORT_NAME} > /dev/null
		check_errs $? "Unable to remove temporary performance report ${CWD}/${REPORT_NAME}/slave${i}/${SLAVE_REPORT_NAME}"
	fi
done

mv ${CWD}/mvn.clean.install.log ${CWD}/${REPORT_NAME}/master/mvn.clean.install.log
check_errs $? "Unable to collect extension log file"

mv ${CWD}/java.PrepareTestPlan.log ${CWD}/${REPORT_NAME}/master/java.PrepareTestPlan.log
check_errs $? "Unable to collect test plan preparation log file"

mv ${CWD}/java.SyncMaster.log ${CWD}/${REPORT_NAME}/master/java.SyncMaster.log
check_errs $? "Unable to collect master synchronization log file"

cp ${HOSTS} ${CWD}/${REPORT_NAME}/master/hosts.properties
check_errs $? "Unable to collect hosts file"

mv ${CWD}/mvn.verify.log ${CWD}/${REPORT_NAME}/master/mvn.verify.log
check_errs $? "Unable to collect performance log file"

mv ${CWD}/tests/${COMPONENT}/target/jmeter/logs/${TESTPLAN_NAME}.log ${CWD}/${REPORT_NAME}/master/
check_errs $? "Unable to collect JMeter log file"

mv ${CWD}/tests/${COMPONENT}/src/test/jmeter/${TESTPLAN_NAME} ${CWD}/${REPORT_NAME}/master/
check_errs $? "Unable to collect the test plan"

mv ${CWD}/tests/${COMPONENT}/target/jmeter/results/*.jtl ${CWD}/${REPORT_NAME}/master/
check_errs $? "Unable to collect JMeter result file"

if [[ -f ${CWD}/tests/${COMPONENT}/summary.jtl ]]; then
	mv ${CWD}/tests/${COMPONENT}/summary.jtl ${CWD}/${REPORT_NAME}/master/summary.jtl
	check_errs $? "Unable to collect the summary report"
fi

if [[ -f ${CWD}/tests/${COMPONENT}/summary.jtl.throughput ]]; then
	mv ${CWD}/tests/${COMPONENT}/summary.jtl.throughput ${CWD}/${REPORT_NAME}/master/summary.jtl.throughput
	check_errs $? "Unable to collect the throughput report"
fi

if [[ -f ${CWD}/tests/${COMPONENT}/RegistryResponseValidation.log ]]; then
	mv ${CWD}/tests/${COMPONENT}/RegistryResponseValidation.log ${CWD}/${REPORT_NAME}/master/RegistryResponseValidation.log
	check_errs $? "Unable to collect RegistryResponseValidation.log"
fi

if [[ -f ${CWD}/tests/${COMPONENT}/RepositoryResponseValidation.log ]]; then
	mv ${CWD}/tests/${COMPONENT}/RepositoryResponseValidation.log ${CWD}/${REPORT_NAME}/master/RepositoryResponseValidation.log
	check_errs $? "Unable to collect RepositoryResponseValidation.log"
fi

if [[ -f ${CWD}/background.start.log ]]; then
	mv ${CWD}/background.start.log ${CWD}/${REPORT_NAME}/master/background.start.log
	check_errs $? "Unable to collect background.start.log"
fi

if [[ -f ${CWD}/background.stop.log ]]; then
	mv ${CWD}/background.stop.log ${CWD}/${REPORT_NAME}/master/background.stop.log
	check_errs $? "Unable to collect background.stop.log"
fi

pushd ${CWD}/${REPORT_NAME} > /dev/null

tar -czf ${CWD}/${REPORT_NAME}.tar.gz . > /dev/null
check_errs $? "Unable to create performance report from ${REPORT_NAME}"

ERRORPERCENTAGE=`find . -name "summary.jtl.throughput" | xargs -n1 head -1 | awk -v NODES=${NODES} '{ERRORPERCENTAGE += $1}END{print ERRORPERCENTAGE/(NODES+1)}'`
THROUGHPUT=`find . -name "summary.jtl.throughput" | xargs -n1 tail -1 | awk '{THROUGHPUT += $1}END{print THROUGHPUT}'`

popd > /dev/null

rm -rf ${CWD}/${REPORT_NAME} > /dev/null
check_errs $? "Unable to clean report directory ${REPORT_NAME}"

pushd ${CWD}/tests/${COMPONENT} > /dev/null

${M3}/mvn clean > /dev/null
check_errs $? "Unable to clean working directories"

popd > /dev/null

echo "Performance test of NSP component '${COMPONENT}' completed."
echo "The report is available in the file ${CWD}/${REPORT_NAME}.tar.gz"

if [[ -n "${THROUGHPUT}" ]]; then
	echo "Throughput was measured to ${THROUGHPUT} requests per second."
	echo "${THROUGHPUT} ${CWD}/${REPORT_NAME}.tar.gz" >> master.throughput
else
	echo "0" >> master.throughput
fi

if [[ -n "${ERRORPERCENTAGE}" ]] && [[ "0" != "${ERRORPERCENTAGE}" ]]; then
	echo "Note: ${ERRORPERCENTAGE}% of the samples failed."
	echo "${ERRORPERCENTAGE} ${CWD}/${REPORT_NAME}.tar.gz" >> master.errorpercentage
else
	echo "0" >> master.errorpercentage
fi

popd > /dev/null
