package com.arosii.jmeter.protocol.ihe.sampler;

import java.nio.charset.StandardCharsets;
import java.util.Calendar;
import java.util.Date;
import java.util.UUID;

import dk.s4.hl7.cda.codes.MedCom;
import dk.s4.hl7.cda.convert.APDXmlCodec;
import dk.s4.hl7.cda.convert.APDXmlCodec.Version;
import dk.s4.hl7.cda.model.AddressData;
import dk.s4.hl7.cda.model.OrganizationIdentity;
import dk.s4.hl7.cda.model.Participant;
import dk.s4.hl7.cda.model.Patient;
import dk.s4.hl7.cda.model.PersonIdentity;
import dk.s4.hl7.cda.model.AddressData.Use;
import dk.s4.hl7.cda.model.OrganizationIdentity.OrganizationBuilder;
import dk.s4.hl7.cda.model.Participant.ParticipantBuilder;
import dk.s4.hl7.cda.model.Patient.Gender;
import dk.s4.hl7.cda.model.PersonIdentity.PersonBuilder;
import dk.s4.hl7.cda.model.apd.AppointmentDocument;
import dk.s4.hl7.cda.model.apd.StatusV10;
import dk.s4.hl7.cda.model.util.DateUtil;

public class OpretAftaleSampler extends ITI41Sampler {

	private static final java.util.Random RANDOM = new java.util.Random();

	private static APDXmlCodec codec = new APDXmlCodec(Version.V10);

	@Override
	protected byte[] getDocumentData() {
		AppointmentDocument appointmentDocument = createAppointmentDocument();
		String xml = codec.encode(appointmentDocument);
		return xml.getBytes(StandardCharsets.UTF_8);
	}

	private static AppointmentDocument createAppointmentDocument() {
		Date time = DateUtil.makeDanishDateTime(2017, 1, 16, 10, 0, 0);
		AppointmentDocument appointment = new AppointmentDocument(MedCom.createId(UUID.randomUUID().toString()));

		appointment.setLanguageCode("da-DK");
		// Title
		appointment.setTitle(text(between(10, 30)));
		appointment.setEffectiveTime(DateUtil.makeDanishDateTime(2017, 0, 13, 10, 0, 0));
		// Patient CPR
		appointment.setPatient(createPatient(createAddress()));
		appointment.setCustodian(createOrganizationIdentity(createAddress()));
		appointment.setAuthor(createParticipant(time, createPersonIdentity(), createOrganizationIdentity(createAddress())));
		// Service Start/Stop Time
		appointment.setDocumentationTimeInterval(DateUtil.makeDanishDateTime(2017, 4, 31, 11, 0, 0), DateUtil.makeDanishDateTime(2017, 4, 31, 12, 0, 0));
		appointment.setAppointmentTitle("Aftale");
		appointment.setAppointmentText(paragraph(between(200, 300)));
		appointment.setAppointmentId(UUID.randomUUID().toString());
		appointment.setAppointmentStatus(StatusV10.ACTIVE);
		appointment.setIndicationDisplayName(paragraph(between(100, 200)));
		appointment.setAppointmentLocation(createOrganizationIdentity(createAddress()));
		appointment.setAppointmentAuthor(createParticipant(time, createPersonIdentity(), createOrganizationIdentity(createAddress())));
		appointment.setAppointmentPerformer(createParticipant(time, createPersonIdentity(), createOrganizationIdentity(createAddress())));

		return appointment;
	}

	/* ---------------- */

	private static Patient createPatient(AddressData address) {
		Patient.PatientBuilder b = new Patient.PatientBuilder();
		b.setGender(gender());
		b.setBirthTime(between(1950, 2010), between(Calendar.JANUARY, Calendar.DECEMBER + 1), between(1, 25));
		b.addFamilyName(text(between(15, 25)));
		b.addGivenName(text(between(5, 15)));
		b.addGivenName(text(between(5, 10)));
		b.setSSN(number(10));
		b.setAddress(address);
		b.addTelecom(AddressData.Use.HomeAddress, "tel", number(8));
		b.addTelecom(AddressData.Use.WorkPlace, "mailto", mail(14));
		return b.build();
	}

	private static Participant createParticipant(Date time, PersonIdentity personIdentity, OrganizationIdentity organizationIdentity) {
		ParticipantBuilder b = new ParticipantBuilder();
		b.setAddress(organizationIdentity.getAddress());
		b.setSOR(organizationIdentity.getIdValue());
		b.setTelecomList(organizationIdentity.getTelecomList());
		b.setTime(time);
		b.setPersonIdentity(personIdentity);
		b.setOrganizationIdentity(organizationIdentity);
		return b.build();
	}

	private static AddressData createAddress() {
		AddressData.AddressBuilder b = new AddressData.AddressBuilder();
		b.addAddressLine(text(between(10, 30)));
		b.addAddressLine(text(between(10, 20)) + " " + between(1, 100));
		b.setPostalCode(String.valueOf(between(1000, 9999)));
		b.setCity(text(between(10, 15)));
		b.setCountry(text(between(5, 10)));
		b.setUse(AddressData.Use.WorkPlace);
		return b.build();
	}

	private static OrganizationIdentity createOrganizationIdentity(AddressData address) {
		OrganizationBuilder b = new OrganizationBuilder();
		b.setSOR(number(16));
		b.setName(text(between(14, 19)));
		b.setAddress(address);
		b.addTelecom(Use.WorkPlace, "tel", number(8));
		return b.build();
	}

	private static PersonIdentity createPersonIdentity() {
		PersonBuilder b = new PersonBuilder();
		b.addFamilyName(text(between(14, 19)));
		b.addGivenName(text(between(10, 15)));
		b.setPrefix(text(between(14, 19)));
		return b.build();
	}

	private static Gender gender() {
		return RANDOM.nextBoolean() ? Gender.Male : Gender.Female;
	}

	private static int between(int min, int max) {
		return RANDOM.nextInt(max - min) + min;
	}

	private static String text(int length) {
		StringBuilder sb = new StringBuilder(length);
		sb.append((char) between('A', 'Z' + 1));
		for (int i = 1; i < length; i++) {
			sb.append((char) between('a', 'z' + 1));
		}
		return sb.toString();
	}

	private static String number(int length) {
		StringBuilder sb = new StringBuilder(length);
		for (int i = 0; i < length; i++) {
			sb.append((char) between('0', '9' + 1));
		}
		return sb.toString();
	}

	private static String mail(int length) {
		StringBuilder sb = new StringBuilder(length + 6);
		for (int i = 0; i < length; i++) {
			if (i == length / 2) {
				sb.append('@');
			}
			sb.append((char) between('a', 'z' + 1));
		}
		sb.append(".test");
		return sb.toString();
	}

	private static String paragraph(int length) {
		StringBuilder sb = new StringBuilder();
		sb.append(text(between(7, 15)));
		while (sb.length() < length) {
			sb.append(' ');
			for (int i = 1; i < between(3, 10); i++) {
				sb.append((char) between('a', 'z' + 1));
			}
		}
		return sb.substring(0, length);
	}
}
