/*
 * The contents of this file is the CONFIDENTAL and PROPRIETARY
 * property of Arosii Information Systems A/S. Any unauthorized use,
 * reproduction or transfer of this file is strictly prohibited.
 *
 * Copyright (C) 2012 Arosii Information Systems A/S.  All rights reserved.
 * This is an unpublished work, and is subject to limited
 * distribution and restricted disclosure only.
 */
package com.arosii.jmeter.protocol.gw.sampler;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.security.GeneralSecurityException;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.util.Enumeration;

import javax.xml.bind.JAXBException;

import org.apache.jmeter.protocol.http.sampler.HTTPSampleResult;
import org.w3._2000._09.xmldsig.KeyInfo;
import org.w3._2000._09.xmldsig.X509Data;
import org.xmlsoap.schemas.soap.envelope.Body;
import org.xmlsoap.schemas.soap.envelope.Envelope;

import com.arosii.jmeter.protocol.sts.sampler.SecurityTokenRequestSampler;

import dk.sosi.gw._2007_09.RequestIdCardDigestForSigningResponse;
import dk.sosi.gw._2007_09.SignIdCardRequestBody;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.SignatureUtil;

/**
 * 
 * @author Jacob Qvortrup
 */
public class GatewaySignIdCardRequestSampler extends AbstractGWSoapSampler {

	private static final long serialVersionUID = 1L;

	public GatewaySignIdCardRequestSampler() {
		setPath("/sosigw/service/sosigw");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected HTTPSampleResult sample(URL url, String method, boolean areFollowingRedirect, int frameDepth) {
		setXmlData(createRequest());
		HTTPSampleResult sample = super.sample(url, method, areFollowingRedirect, frameDepth);
		return sample;
	}

	private String createRequest() {
		try {
			RequestIdCardDigestForSigningResponse requestIdCardDigestForSigningResponse = getRequestIdCardDigestForSigningResponse();

			Envelope envelope = new Envelope();
			envelope.setHeader(makeIdCardInSecurityHeader());
			envelope.setBody(makeSignIdCardRequestBody(requestIdCardDigestForSigningResponse));

			return marshall(envelope);
		} catch (JAXBException e) {
			throw new RuntimeException(e);
		}
	}

	private Body makeSignIdCardRequestBody(RequestIdCardDigestForSigningResponse requestIdCardDigestForSigningResponse) {
		try {
			SignIdCardRequestBody signIdCardRequestBody = signIdCardDigest(requestIdCardDigestForSigningResponse);
			Body body = new Body();
			body.getAny().add(signIdCardRequestBody);
			return body;
		} catch (GeneralSecurityException e) {
			throw new RuntimeException(e);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private SignIdCardRequestBody signIdCardDigest(RequestIdCardDigestForSigningResponse response) throws GeneralSecurityException, IOException {
		byte[] digestValue = response.getDigestValue();
		String keystorePath = getPropertyAsString(SecurityTokenRequestSampler.CERTIFICATE_PATH);
		String keystorePassword = getPropertyAsString(SecurityTokenRequestSampler.CERTIFICATE_PASSWORD);

		KeyStore keystore;
		if (keystorePath.endsWith(".pkcs12") || keystorePath.endsWith(".p12")) {
			keystore = KeyStore.getInstance("PKCS12", SignatureUtil.getCryptoProvider(SignatureUtil.setupCryptoProviderForJVM(), SOSIFactory.PROPERTYNAME_SOSI_CRYPTOPROVIDER_PKCS12));
		} else {
			keystore = KeyStore.getInstance("JKS");
		}
		InputStream keyStoreInputStream = null;
		try {
			if (new File(keystorePath).isFile()) {
				keyStoreInputStream = new FileInputStream(keystorePath);
			} else {
				keyStoreInputStream = Thread.currentThread().getContextClassLoader().getResourceAsStream(keystorePath);
			}
			keystore.load(keyStoreInputStream, keystorePassword.toCharArray());
		} finally {
			try {
				keyStoreInputStream.close();
			} catch (Exception ignore) {
			}
		}

		Certificate certificate = keystore.getCertificate("SOSI:ALIAS_SYSTEM");
		PrivateKey key = (PrivateKey) keystore.getKey("SOSI:ALIAS_SYSTEM", keystorePassword.toCharArray());

		if (certificate == null && key == null) {
			Enumeration<String> aliases = keystore.aliases();
			while (aliases.hasMoreElements()) {
				String alias = aliases.nextElement();
				if (keystore.isKeyEntry(alias)) {
					certificate = (X509Certificate) keystore.getCertificate(alias);
					key = (PrivateKey) keystore.getKey(alias, keystorePassword.toCharArray());
				}
			}
		}

		Signature signature;
		signature = Signature.getInstance("SHA1withRSA");
		signature.initSign(key);
		signature.update(digestValue);
		byte[] signatureValue = signature.sign();

		byte[] encodedCertificate = certificate.getEncoded();
		X509Data x509Data = new X509Data();
		x509Data.setX509Certificate(encodedCertificate);
		KeyInfo keyInfo = new KeyInfo();
		keyInfo.setX509Data(x509Data);

		SignIdCardRequestBody signIdCardRequestBody = new SignIdCardRequestBody();
		signIdCardRequestBody.setSignatureValue(signatureValue);
		signIdCardRequestBody.setKeyInfo(keyInfo);
		return signIdCardRequestBody;
	}
}
