package com.arosii.jmeter.protocol.ihe.cxf;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.annotation.XmlEnumValue;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.cxf.binding.soap.SoapMessage;
import org.apache.cxf.binding.soap.interceptor.AbstractSoapInterceptor;
import org.apache.cxf.headers.Header;
import org.apache.cxf.interceptor.Fault;
import org.apache.cxf.phase.Phase;
import org.apache.jmeter.samplers.Sampler;
import org.apache.jmeter.threads.JMeterContext;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import com.arosii.jmeter.protocol.ihe.sampler.AbstractIHESampler;

import dk.nsi.hsuid.ActingUserCivilRegistrationNumberAttribute;
import dk.nsi.hsuid.Attribute;
import dk.nsi.hsuid.CitizenCivilRegistrationNumberAttribute;
import dk.nsi.hsuid.ConsentOverrideAttribute;
import dk.nsi.hsuid.HealthcareServiceUserIdentificationHeaderUtil;
import dk.nsi.hsuid.OperationsOrganisationNameAttribute;
import dk.nsi.hsuid.OrganisationIdentifierAttribute;
import dk.nsi.hsuid.ResponsibleUserAuthorizationCodeAttribute;
import dk.nsi.hsuid.ResponsibleUserCivilRegistrationNumberAttribute;
import dk.nsi.hsuid.SystemNameAttribute;
import dk.nsi.hsuid.SystemVendorNameAttribute;
import dk.nsi.hsuid.SystemVersionAttribute;
import dk.nsi.hsuid.UserTypeAttribute;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import dk.nsi.hsuid._2016._08.hsuid_1_1.SubjectIdentifierType;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.UserIDCard;

public class HSUIDInterceptor extends AbstractSoapInterceptor {

	private static final DocumentBuilderFactory DOCUMENT_BUILDER_FACTORY = DocumentBuilderFactory.newInstance();

	private static final JAXBContext JAXB_CONTEXT;

	static {
		try {
			JAXB_CONTEXT = JAXBContext.newInstance(HsuidHeader.class);
		} catch (JAXBException e) {
			throw new IllegalStateException(e);
		}
	}

	private final JMeterContext jMeterContext;

	public HSUIDInterceptor(JMeterContext jMeterContext) {
		super(Phase.PRE_STREAM);
		this.jMeterContext = jMeterContext;
	}

	@Override
	public void handleMessage(SoapMessage message) throws Fault {
		UserIDCard idCard = (UserIDCard) jMeterContext.getVariables().getObject(IDCard.class.getName());

		Sampler currentSampler = jMeterContext.getCurrentSampler();
		if (currentSampler instanceof AbstractIHESampler<?>) {
			AbstractIHESampler<?> iheSampler = (AbstractIHESampler<?>) currentSampler;
			String citizenCivilRegistrationNumber = iheSampler.getCitizenCPRNumber();
			String organisationIdentifier = iheSampler.getOrganisationIdentifier();
			String organisationIdentifierFormat = iheSampler.getOrganisationIdentifierFormat();

			List<Attribute> attributes = new ArrayList<Attribute>();
			attributes.add(UserTypeAttribute.HEALTHCAREPROFESSIONAL);
			attributes.add(new CitizenCivilRegistrationNumberAttribute(citizenCivilRegistrationNumber));
			attributes.add(new ActingUserCivilRegistrationNumberAttribute(idCard.getUserInfo().getCPR()));
			attributes.add(new OrganisationIdentifierAttribute(organisationIdentifier, getFormat(organisationIdentifierFormat)));
			attributes.add(new SystemVendorNameAttribute("Sundhedsdatastyrelsen"));
			attributes.add(new SystemNameAttribute(idCard.getSystemInfo().getITSystemName()));
			attributes.add(new SystemVersionAttribute("1.0.0"));
			attributes.add(new OperationsOrganisationNameAttribute(idCard.getSystemInfo().getCareProvider().getOrgName()));
			attributes.add(new ConsentOverrideAttribute(true));
			attributes.add(new ResponsibleUserCivilRegistrationNumberAttribute(idCard.getUserInfo().getCPR()));
			attributes.add(new ResponsibleUserAuthorizationCodeAttribute(idCard.getUserInfo().getAuthorizationCode()));
			HsuidHeader hsuidHeader = HealthcareServiceUserIdentificationHeaderUtil.createHealthcareServiceUserIdentificationHeader("NSP Performance Framework", attributes);

			Document document;
			try {
				document = DOCUMENT_BUILDER_FACTORY.newDocumentBuilder().newDocument();
				JAXB_CONTEXT.createMarshaller().marshal(hsuidHeader, document);
			} catch (ParserConfigurationException e) {
				throw new IllegalStateException(e);
			} catch (JAXBException e) {
				throw new IllegalStateException(e);
			}

			Node hsuidElement = document.getDocumentElement().getFirstChild();
			Header header = new Header(new QName(hsuidElement.getNamespaceURI(), hsuidElement.getLocalName()), document.getFirstChild());
			message.getHeaders().add(header);
		} else {
			throw new ClassCastException(currentSampler.getClass().getName() + " not instance of " + AbstractIHESampler.class.getName());
		}
	}

	private String getFormat(String organisationIdentifierFormat) {
		for (SubjectIdentifierType subjectIdentifierType : SubjectIdentifierType.values()) {
			try {
				if (SubjectIdentifierType.class.getField(subjectIdentifierType.name()).getAnnotation(XmlEnumValue.class).value().equals(organisationIdentifierFormat)) {
					return subjectIdentifierType.toString();
				}
			} catch (Exception e) {
				throw new IllegalStateException(e);
			}
		}
		throw new IllegalArgumentException(organisationIdentifierFormat);
	}

}
