package com.arosii.jmeter.protocol.ihe.sampler;

import static org.apache.jmeter.protocol.http.sampler.HTTPSamplerBase.DOMAIN;
import static org.apache.jmeter.protocol.http.sampler.HTTPSamplerBase.PATH;
import static org.apache.jmeter.protocol.http.sampler.HTTPSamplerBase.PORT;
import static org.apache.jmeter.protocol.http.sampler.HTTPSamplerBase.PROTOCOL;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import javax.xml.ws.soap.SOAPFaultException;

import org.apache.cxf.endpoint.Client;
import org.apache.cxf.frontend.ClientProxy;
import org.apache.jmeter.protocol.http.util.HTTPConstants;
import org.apache.jmeter.samplers.AbstractSampler;
import org.apache.jmeter.samplers.Entry;
import org.apache.jmeter.samplers.SampleResult;
import org.apache.jorphan.util.JOrphanUtils;
import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;
import org.joda.time.format.ISODateTimeFormat;
import org.openehealth.ipf.commons.ihe.ws.WsTransactionConfiguration;
import org.openehealth.ipf.commons.ihe.xds.core.XdsClientFactory;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Status;

import com.arosii.jmeter.protocol.ihe.cxf.ContentCollectorInInterceptor;
import com.arosii.jmeter.protocol.ihe.cxf.ContentCollectorOutInterceptor;
import com.arosii.jmeter.protocol.ihe.cxf.DGWSInterceptor;
import com.arosii.jmeter.protocol.ihe.cxf.HSUIDInterceptor;
import com.arosii.jmeter.protocol.ihe.io.ContentCollectorInputStream;
import com.arosii.jmeter.protocol.ihe.io.ContentCollectorOutputStream;

public abstract class AbstractIHESampler<T> extends AbstractSampler {

	public static final String ADD_DGWS_HEADERS = "IHESampler.add_dgws_header";

	public static final String CITIZEN_CPR_NUMBER = "IHESampler.cpr";

	public static final String ORGANISATION_NAME = "IHESampler.organisation_name";

	public static final String ORGANISATION_IDENTIFIER = "IHESampler.organisation_identifier";

	public static final String ORGANISATION_IDENTIFIER_FORMAT = "IHESampler.organisation_identifier_format";

	public boolean useDGWSHeaders() {
		return getPropertyAsBoolean(ADD_DGWS_HEADERS, false);
	}

	public String getCitizenCPRNumber() {
		return getPropertyAsString(CITIZEN_CPR_NUMBER);
	}

	public String getOrganisationName() {
		return getPropertyAsString(ORGANISATION_NAME);
	}

	public String getOrganisationIdentifier() {
		return getPropertyAsString(ORGANISATION_IDENTIFIER);
	}

	public String getOrganisationIdentifierFormat() {
		return getPropertyAsString(ORGANISATION_IDENTIFIER_FORMAT);
	}

	public String getProtocol() {
		String protocol = getPropertyAsString(PROTOCOL);
		if (protocol == null || protocol.length() == 0) {
			return HTTPConstants.PROTOCOL_HTTP;
		}
		return protocol;
	}

	public String getDomain() {
		return getPropertyAsString(DOMAIN);
	}

	public int getPort() {
		return getPropertyAsInt(PORT, 0);
	}

	public String getPath() {
		String path = JOrphanUtils.replaceAllChars(getPropertyAsString(PATH), ' ', "%20");
		if (path == null || path.trim().isEmpty()) {
			return getPath(useDGWSHeaders());
		} else {
			return path;
		}
	}

	protected abstract String getPath(boolean usingDGWS);

	public URL getUrl() {
		try {
			return new URL(getProtocol(), getDomain(), getPort(), getPath());
		} catch (MalformedURLException e) {
			throw new IllegalArgumentException(e);
		}
	}

	@Override
	public SampleResult sample(Entry entry) {
		SampleResult res = null;
		try {
			res = sample();
			res.setSampleLabel(getName());
			return res;
		} catch (Exception e) {
			return errorResult(e);
		}
	}

	protected abstract SampleResult sample() throws Exception;

	protected IHESampleResult errorResult(Throwable t) {
		IHESampleResult res = new IHESampleResult();
		res.setSampleLabel(t.getClass().getSimpleName() + ": " + getName());
		res.setDataType(SampleResult.TEXT);
		if (t instanceof SOAPFaultException) {
			try {
				res.setSamplerData(getContentCollectorOutputStream().content());
			} catch (Exception e) {
				ByteArrayOutputStream text = new ByteArrayOutputStream(200);
				e.printStackTrace(new PrintStream(text));
				try {
					res.setSamplerData(new String(text.toByteArray(), "utf-8"));
				} catch (UnsupportedEncodingException ignored) {
					res.setSamplerData(new String(text.toByteArray()));
				}
			}
		}
		ByteArrayOutputStream text = new ByteArrayOutputStream(200);
		t.printStackTrace(new PrintStream(text));
		res.setResponseData(text.toByteArray());
		res.setResponseCode(t.getClass().getName());
		res.setResponseMessage(t.getMessage());
		res.setSuccessful(false);
		res.setMonitor(false);
		return res;
	}

	protected void updateSampleResult(IHESampleResult res, Status status) throws IOException {
		res.setURL(getUrl());
		res.setResponseMessage(status.getOpcode30());
		res.setSamplerData(getContentCollectorOutputStream().content());
		res.setResponseData(getContentCollectorInputStream().content(), "utf-8");
		res.setDataEncoding("utf-8");
		res.setContentType("text/xml");
		switch (status) {
		case SUCCESS:
			res.setResponseCode(Integer.toString(HttpURLConnection.HTTP_OK));
			res.setSuccessful(true);
			break;
		case PARTIAL_SUCCESS:
			res.setResponseCode(Integer.toString(HttpURLConnection.HTTP_OK));
			res.setSuccessful(false);
			break;
		case FAILURE:
			res.setResponseCode(Integer.toString(HttpURLConnection.HTTP_INTERNAL_ERROR));
			res.setSuccessful(false);
			break;
		}
	}

	@SuppressWarnings("unchecked")
	protected T getClient(Class<T> clazz) {
		URL url = getUrl();
		boolean dgws = useDGWSHeaders();

		Map<String, Object> clients = (Map<String, Object>) getThreadContext().getVariables().getObject(clazz.getName());
		if (clients == null) {
			clients = new HashMap<>();
			getThreadContext().getVariables().putObject(clazz.getName(), clients);
		}

		String key = String.valueOf(url) + String.valueOf(dgws);
		T client = (T) clients.get(key);
		if (client == null) {
			client = createClient(url, clazz);
			clients.put(key, client);
		}
		return client;
	}

	@SuppressWarnings("unchecked")
	protected T createClient(URL url, Class<T> clazz) {
		String wsdl = "wsdl/" + clazz.getSimpleName().substring(0, 5).toLowerCase() + ".wsdl";
		WsTransactionConfiguration configuration = getWsTransactionConfiguration(wsdl, clazz);
		T client = (T) new XdsClientFactory(configuration, url.toExternalForm(), null, null, null).getClient();
		Client cxfClient = ClientProxy.getClient(client);
		cxfClient.getInInterceptors().add(new ContentCollectorInInterceptor(getThreadContext()));
		cxfClient.getOutInterceptors().add(new ContentCollectorOutInterceptor(getThreadContext()));
		if (useDGWSHeaders()) {
			cxfClient.getOutInterceptors().add(new DGWSInterceptor(getThreadContext()));
			cxfClient.getOutInterceptors().add(new HSUIDInterceptor(getThreadContext()));
		}
		return client;
	}

	protected abstract WsTransactionConfiguration getWsTransactionConfiguration(String wsdl, Class<T> clazz);

	protected ContentCollectorOutputStream getContentCollectorOutputStream() {
		return (ContentCollectorOutputStream) getThreadContext().getVariables().getObject(ContentCollectorOutputStream.class.getName());
	}

	protected ContentCollectorInputStream getContentCollectorInputStream() {
		return (ContentCollectorInputStream) getThreadContext().getVariables().getObject(ContentCollectorInputStream.class.getName());
	}

	protected DateTime getDateTime(String property) {
		String value = getPropertyAsString(property, "");
		if (value.trim().isEmpty()) {
			return null;
		} else {
			return ISODateTimeFormat.dateTimeParser().withZone(DateTimeZone.getDefault()).parseDateTime(value);
		}
	}
}
