package com.arosii.jmeter.protocol.ihe.sampler;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.Set;

import org.apache.jmeter.config.Argument;
import org.apache.jmeter.config.Arguments;
import org.apache.jmeter.samplers.SampleResult;
import org.apache.jmeter.testelement.property.PropertyIterator;
import org.joda.time.DateTime;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.EbXMLAdhocQueryRequest;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.ebxml30.EbXMLFactory30;
import org.openehealth.ipf.commons.ihe.xds.core.ebxml.ebxml30.EbXMLQueryResponse30;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AssigningAuthority;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.AvailabilityStatus;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntryType;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Identifiable;
import org.openehealth.ipf.commons.ihe.xds.core.requests.QueryRegistry;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.FindDocumentsQuery;
import org.openehealth.ipf.commons.ihe.xds.core.requests.query.QueryReturnType;
import org.openehealth.ipf.commons.ihe.xds.core.responses.QueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryRequest;
import org.openehealth.ipf.commons.ihe.xds.core.stub.ebrs30.query.AdhocQueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.transform.requests.QueryRegistryTransformer;
import org.openehealth.ipf.commons.ihe.xds.core.transform.responses.QueryResponseTransformer;
import org.openehealth.ipf.commons.ihe.xds.iti18.Iti18PortType;

public class ITI18Sampler extends AbstractIHEDocumentRegistrySampler<Iti18PortType> {

	public static final String USE_CITIZEN_CPR_NUMBER_FILE = "ITI18Sampler.use_cpr_file";

	public static final String CITIZEN_CPR_NUMBER_FILE_NAME = "ITI18Sampler.cpr_file_name";

	public static final String TYPE_CODES = "ITI18Sampler.type_codes";

	public static final String FORMAT_CODES = "ITI18Sampler.format_codes";

	public static final String STABLE = "ITI18Sampler.stable";

	public static final String ON_DEMAND = "ITI18Sampler.on_demand";

	public static final String SERVICE_START_TIME_FROM = "ITI18Sampler.service_start_time_from";

	public static final String SERVICE_START_TIME_TO = "ITI18Sampler.service_start_time_to";

	public static final Random RANDOM = new Random();

	protected static List<String> cprCache;

	protected static EbXMLFactory30 ebXMLFactory30 = new EbXMLFactory30();

	protected static QueryRegistryTransformer queryRegistryTransformer = new QueryRegistryTransformer();

	protected static QueryResponseTransformer queryResponseTransformer = new QueryResponseTransformer(ebXMLFactory30);

	@Override
	protected String getPath(boolean usingDGWS) {
		if (usingDGWS) {
			return "/ddsregistry/service";
		} else {
			return "/registry/services/xds-iti18";
		}
	}

	@Override
	public String getCitizenCPRNumber() {
		String cpr = getThreadContext().getVariables().get(ITI18Sampler.class.getName() + ".cpr");
		if (cpr == null || cpr.trim().isEmpty()) {
			if (getPropertyAsBoolean(USE_CITIZEN_CPR_NUMBER_FILE, false)) {
				cpr = getRandomCitizenCPRNumber(getPropertyAsString(CITIZEN_CPR_NUMBER_FILE_NAME, getClass().getSimpleName() + ".cpr"));
			} else {
				cpr = super.getCitizenCPRNumber();
			}
			getThreadContext().getVariables().put(ITI18Sampler.class.getName() + ".cpr", cpr);
		}
		return cpr;
	}

	@Override
	protected SampleResult sample() throws Exception {
		getThreadContext().getVariables().remove(ITI18Sampler.class.getName() + ".cpr");
		ITI18SampleResult res = new ITI18SampleResult();
		res.setHTTPMethod("ITI18");
		Iti18PortType iti18Client = getClient(Iti18PortType.class);
		QueryRegistry queryRegistry = buildQueryRegistry();
		res.setQueryRegistry(queryRegistry);

		AdhocQueryRequest adhocQueryRequest = marshall(queryRegistry);
		res.sampleStart();
		AdhocQueryResponse adhocQueryResponse;
		adhocQueryResponse = iti18Client.documentRegistryRegistryStoredQuery(adhocQueryRequest);
		res.sampleEnd();
		QueryResponse queryResponse = unmarshall(adhocQueryResponse);
		res.setQueryResponse(queryResponse);

		updateSampleResult(res, queryResponse.getStatus());
		return res;
	}

	protected QueryRegistry buildQueryRegistry() {
		FindDocumentsQuery findDocumentsQuery = new FindDocumentsQuery();

		findDocumentsQuery.setPatientId(new Identifiable(getCitizenCPRNumber(), getCitizenAssigningAuthority()));

		findDocumentsQuery.setStatus(getAvailabilityStatus());

		findDocumentsQuery.setDocumentEntryTypes(getDocumentEntryTypes());

		findDocumentsQuery.setTypeCodes(getTypeCodes());

		findDocumentsQuery.setFormatCodes(getFormatCodes());

		findDocumentsQuery.getServiceStartTime().setFrom(getServiceStartTimeFrom());

		findDocumentsQuery.getServiceStartTime().setTo(getServiceStartTimeTo());

		QueryRegistry queryRegistry = new QueryRegistry(findDocumentsQuery);
		queryRegistry.setReturnType(QueryReturnType.LEAF_CLASS);

		return queryRegistry;
	}

	protected AssigningAuthority getCitizenAssigningAuthority() {
		return new AssigningAuthority("1.2.208.176.1.2");
	}

	protected DateTime getServiceStartTimeTo() {
		return getDateTime(SERVICE_START_TIME_TO);
	}

	protected DateTime getServiceStartTimeFrom() {
		return getDateTime(SERVICE_START_TIME_FROM);
	}

	protected List<Code> getFormatCodes() {
		return getCodes(FORMAT_CODES);
	}

	protected List<Code> getTypeCodes() {
		return getCodes(TYPE_CODES);
	}

	protected List<AvailabilityStatus> getAvailabilityStatus() {
		return Arrays.asList(AvailabilityStatus.APPROVED);
	}

	protected List<Code> getCodes(String property) {
		List<Code> codes = new LinkedList<>();
		Arguments arguments = (Arguments) getProperty(property).getObjectValue();
		if (arguments != null) {
			for (PropertyIterator iterator = arguments.iterator(); iterator.hasNext();) {
				Argument argument = (Argument) iterator.next().getObjectValue();
				// 2.16.840.1.113883.6.1 means LOINC codes
				codes.add(new Code(argument.getName(), null, argument.getValue()));
			}
		}
		return codes;
	}

	protected List<DocumentEntryType> getDocumentEntryTypes() {
		List<DocumentEntryType> documentEntryTypes = new ArrayList<>(2);
		if (getPropertyAsBoolean(STABLE, true)) {
			documentEntryTypes.add(DocumentEntryType.STABLE);
		}
		if (getPropertyAsBoolean(ON_DEMAND, false)) {
			documentEntryTypes.add(DocumentEntryType.ON_DEMAND);
		}
		return documentEntryTypes;
	}

	protected static AdhocQueryRequest marshall(QueryRegistry queryRegistry) {
		EbXMLAdhocQueryRequest ebxmlAdhocQueryRequest = queryRegistryTransformer.toEbXML(queryRegistry);
		AdhocQueryRequest adhocQueryRequest = (AdhocQueryRequest) ebxmlAdhocQueryRequest.getInternal();
		return adhocQueryRequest;
	}

	protected static QueryResponse unmarshall(AdhocQueryResponse adhocQueryResponse) {
		EbXMLQueryResponse30 ebXmlresponse = new EbXMLQueryResponse30(adhocQueryResponse);
		QueryResponse queryResponse = queryResponseTransformer.fromEbXML(ebXmlresponse);
		return queryResponse;
	}

	protected String getRandomCitizenCPRNumber(String cprFilename) {
		loadCitizenCPRNumbers(cprFilename);
		String cpr = cprCache.get(RANDOM.nextInt(cprCache.size()));
		return cpr;
	}

	protected synchronized void loadCitizenCPRNumbers(String filename) {
		if (cprCache == null) {
			BufferedReader reader = null;
			try {
				InputStream cprInputStream;
				File cprFile = new File(filename);
				if (cprFile.exists() && cprFile.canRead()) {
					cprInputStream = new BufferedInputStream(new FileInputStream(cprFile));
				} else {
					cprInputStream = new BufferedInputStream(getClass().getResourceAsStream(filename));
				}
				reader = new BufferedReader(new InputStreamReader(cprInputStream, "utf-8"));
				Set<String> cprNumbers = new HashSet<String>();
				for (String line = reader.readLine(); line != null; line = reader.readLine()) {
					line = line.trim();
					if (!line.isEmpty() && !line.startsWith("#")) {
						cprNumbers.add(line);
					}
				}
				cprCache = new ArrayList<String>(cprNumbers);
			} catch (IOException e) {
				throw new IllegalStateException(e);
			} finally {
				if (reader != null) {
					try {
						reader.close();
					} catch (IOException ignore) {
					}
				}
			}
		}
	}
}
