package com.arosii.jmeter.protocol.ihe.visualizers;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.jmeter.samplers.SampleResult;
import org.apache.jmeter.visualizers.ResultRenderer;
import org.apache.jmeter.visualizers.SamplerResultTab;
import org.apache.jmeter.visualizers.ViewResultsFullVisualizer;
import org.joda.time.format.DateTimeFormat;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.Code;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.DocumentEntry;
import org.openehealth.ipf.commons.ihe.xds.core.metadata.LocalizedString;
import org.openehealth.ipf.commons.ihe.xds.core.responses.ErrorInfo;
import org.openehealth.ipf.commons.ihe.xds.core.responses.QueryResponse;
import org.openehealth.ipf.commons.ihe.xds.core.responses.Response;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocument;
import org.openehealth.ipf.commons.ihe.xds.core.responses.RetrievedDocumentSet;

import com.arosii.jmeter.protocol.ihe.sampler.ITI18SampleResult;
import com.arosii.jmeter.protocol.ihe.sampler.ITI41SampleResult;
import com.arosii.jmeter.protocol.ihe.sampler.ITI43DocumentSampleResult;
import com.arosii.jmeter.protocol.ihe.sampler.ITI43SampleResult;
import com.arosii.jmeter.resources.ihe.Messages;

public class RenderAsIHE extends SamplerResultTab implements ResultRenderer {

	/** {@inheritDoc} */
	@Override
	public void renderResult(SampleResult sampleResult) {
		StringBuilder iheText = new StringBuilder();
		String contentType = "text/plain";
		if (sampleResult instanceof ITI18SampleResult) {
			contentType = "text/html";
			iheText.append("<html><body>");

			ITI18SampleResult iti18SampleResult = (ITI18SampleResult) sampleResult;
			QueryResponse queryResponse = iti18SampleResult.getQueryResponse();
			if (queryResponse != null) {
				iheText.append("<h3>").append(queryResponse.getStatus().toString()).append("</h3>");
				for (ErrorInfo error : queryResponse.getErrors()) {
					iheText.append("<p><b>").append(StringEscapeUtils.escapeHtml4(error.toString())).append("</b></p>");
				}
				for (DocumentEntry documentEntry : queryResponse.getDocumentEntries()) {
					iheText.append("<p>");
					if (documentEntry.getTitle() != null && documentEntry.getTitle().getValue() != null && !documentEntry.getTitle().getValue().isEmpty()) {
						iheText.append("<h3>").append(documentEntry.getTitle().getValue()).append("</h3>");
					} else {
						iheText.append("<h3>").append(Messages.getResString("view_results_render_ihe_no_document_title")).append("</h3>");
					}
					iheText.append(Messages.getResString("view_results_render_ihe_service_start_time")).append(" : ").append(DateTimeFormat.forPattern("yyyy-MM-dd HH:mm:ss").print(documentEntry.getServiceStartTime())).append("<br/>");
					iheText.append(Messages.getResString("view_results_render_ihe_service_stop_time")).append(" : ").append(DateTimeFormat.forPattern("yyyy-MM-dd HH:mm:ss").print(documentEntry.getServiceStopTime())).append("<br/>");
					iheText.append(Messages.getResString("view_results_render_ihe_type_code")).append(" : ").append(format(documentEntry.getTypeCode())).append("<br/>");
					iheText.append(Messages.getResString("view_results_render_ihe_format_code")).append(" : ").append(format(documentEntry.getFormatCode())).append("<br/>");
					iheText.append(Messages.getResString("view_results_render_ihe_repository_unique_id")).append(" : ").append(documentEntry.getRepositoryUniqueId()).append("<br/>");
					iheText.append(Messages.getResString("view_results_render_ihe_document_unique_id")).append(" : ").append(documentEntry.getUniqueId());
					iheText.append("</p>");
				}
			} else {
				iheText.append("NULL");
			}
			iheText.append("</body></html>");
		} else if (sampleResult instanceof ITI43SampleResult) {
			contentType = "text/html";
			iheText.append("<html><body>");
			ITI43SampleResult iti43SampleResult = (ITI43SampleResult) sampleResult;
			RetrievedDocumentSet retrievedDocumentSet = iti43SampleResult.getRetrievedDocumentSet();
			if (retrievedDocumentSet != null) {
				iheText.append("<h3>").append(retrievedDocumentSet.getStatus().toString()).append("</h3>");
				for (ErrorInfo error : retrievedDocumentSet.getErrors()) {
					iheText.append("<p><b>").append(StringEscapeUtils.escapeHtml4(error.toString())).append("</b></p>");
				}
				for (RetrievedDocument retrievedDocument : retrievedDocumentSet.getDocuments()) {
					String documentUniqueId = retrievedDocument.getRequestData().getDocumentUniqueId();
					iheText.append("<p>");
					iheText.append("<h3>").append(documentUniqueId).append("</h3>");
					iheText.append(Messages.getResString("view_results_render_ihe_repository_unique_id")).append(" : ").append(retrievedDocument.getRequestData().getRepositoryUniqueId()).append("<br/>");
					iheText.append(Messages.getResString("view_results_render_ihe_document_unique_id")).append(" : ").append(documentUniqueId).append("<br/>");
					iheText.append(Messages.getResString("view_results_render_ihe_content_type")).append(" : ").append(retrievedDocument.getDataHandler().getContentType()).append("<br/>");
					iheText.append(Messages.getResString("view_results_render_ihe_document_size")).append(" : ").append(iti43SampleResult.getDocument(documentUniqueId).length()).append(' ').append(Messages.getResString("view_results_render_ihe_bytes")).append("<br/>");
					iheText.append("</p>");
				}
			} else {
				iheText.append("NULL");
			}
			iheText.append("</body></html>");
		} else if (sampleResult instanceof ITI43DocumentSampleResult) {
			contentType = "text/plain";
			ITI43DocumentSampleResult iti43DocumentSampleResult = (ITI43DocumentSampleResult) sampleResult;
			iheText.append(iti43DocumentSampleResult.getDocument());
		} else if (sampleResult instanceof ITI41SampleResult) {
			contentType = "text/html";
			iheText.append("<html><body>");
			ITI41SampleResult iti41SampleResult = (ITI41SampleResult) sampleResult;
			Response response = iti41SampleResult.getResponse();
			if (response != null) {
				iheText.append("<h3>").append(response.getStatus().toString()).append("</h3>");
				for (ErrorInfo error : response.getErrors()) {
					iheText.append("<p><b>").append(StringEscapeUtils.escapeHtml4(error.toString())).append("</b></p>");
				}
			} else {
				iheText.append("NULL");
			}
			iheText.append("</body></html>");
		}

		if (iheText.length() > 0) {
			showTextResponse(iheText.toString(), contentType);
		} else {
			showTextResponse(ViewResultsFullVisualizer.getResponseAsString(sampleResult), "text/plain");
		}
	}

	public static String format(Code code) {
		StringBuilder sb = new StringBuilder();
		if (code != null) {
			LocalizedString displayName = code.getDisplayName();
			if (displayName != null && displayName.getValue() != null && !displayName.getValue().isEmpty()) {
				sb.append(displayName.getValue()).append("<br/");
			} else {
				sb.append(code.getCode()).append(" / ").append(code.getSchemeName()).append("<br/");
			}
			sb.append("&nbsp;&nbsp;").append(Messages.getResString("view_results_render_ihe_classification_code")).append(" : ").append(code.getCode()).append("<br/>");
			sb.append("&nbsp;&nbsp;").append(Messages.getResString("view_results_render_ihe_classification_scheme")).append(" : ").append(code.getSchemeName());
		}
		return sb.toString();
	}

	private void showTextResponse(String response, String contentType) {
		results.setContentType(contentType); // $NON-NLS-1$
		results.setText(response == null ? "" : response); // $NON-NLS-1$
		results.setCaretPosition(0);
		resultsScrollPane.setViewportView(results);
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return Messages.getResString("view_results_render_ihe"); // $NON-NLS-1$
	}

}
