/**
* The MIT License
*   
* Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
*   
* Copyright (C) 2018 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
*   
*   
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*   
*   
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*   
*   
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*/ 

package dk.nsp.jmeter.protocol.lar.sampler;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.jmeter.protocol.http.sampler.HTTPSampleResult;
import org.apache.jorphan.logging.LoggingManager;
import org.apache.log.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.arosii.jmeter.protocol.nsp.sampler.AbstractNSPSoapSampler;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.xml.XmlUtil;

/**
 * 
 * @author
 */
public class ListAllergyRequestSampler extends AbstractNSPSoapSampler {

	private static final long serialVersionUID = 1L;
	
	private static final Logger log = LoggingManager.getLoggerForClass();

	public static final String CPR = "ListAllergyRequestSampler.CPR";
	
	public static final String CPR_SYSTEM = "ListAllergyRequestSampler.CPRSystem";

	public static final String CPR_FILE = "ListAllergyRequestSampler.CPRFile";
	
	public static final String ORG_CODE = "ListAllergyRequestSampler.OrgCode";
	
	public static final String ORG_TYPE = "ListAllergyRequestSampler.OrgType";

	public static final String CONSENT_OVERRIDE = "ListAllergyRequestSampler.ConsentOverride";

	public static final Random RANDOM = new Random();

	private static List<String> cprCache;

	public ListAllergyRequestSampler() {
		setPath("/larservice/MedicationAllergyService");
		setSOAPAction("/GetPatientAllergyList");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected HTTPSampleResult sample(URL url, String method, boolean areFollowingRedirect, int frameDepth) {
		clearReply();
		setXmlData(createRequest());
		HTTPSampleResult sample = super.sample(url, method, areFollowingRedirect, frameDepth);
		putReply(sample);
		return sample;
	}

	private String createRequest() {
		SOSIFactory sosiFactory = getSOSIFactory();
		IDCard idCard = getIDCard();

		// Create a new request and add the id card to it.
		Request request = sosiFactory.createNewRequest(false, null);
		request.setIDCard(idCard);

		Document bodyDocument = createListAllergyRequestBody();

		request.setBody(bodyDocument.getDocumentElement());

		Document doc = request.serialize2DOMDocument();

		return XmlUtil.node2String(doc, false, true);
	}

	private Document createListAllergyRequestBody() {
		Document bodyDocument;
		try {
			bodyDocument = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
		} catch (ParserConfigurationException e) {
			throw new IllegalStateException(e);
		}
		String requestMethod = "GetPatientAllergyListRequest";
		Element listAllergyRequest = bodyDocument.createElementNS("http://nsp.dk/medicationallergyservice/2018/11/21/", requestMethod);
		listAllergyRequest.setPrefix("ns6");
		bodyDocument.appendChild(listAllergyRequest);

		String cprProperty = getPropertyAsString(CPR);
		String cprSystemProperty = getPropertyAsString(CPR_SYSTEM);
		String cprFileProperty = getPropertyAsString(CPR_FILE);
		
		String orgCodeProperty = getPropertyAsString(ORG_CODE);
		String orgTypeProperty = getPropertyAsString(ORG_TYPE);
		
		String consetOverride = getPropertyAsString(CONSENT_OVERRIDE, "false");

		if (cprFileProperty != null && !cprFileProperty.trim().isEmpty()) {
			cprProperty = getRandomCPR(cprFileProperty);
		}

		Element medicationAllergyIntoleranceSearchCriteria = bodyDocument.createElement("MedicationAllergyIntoleranceSearchCriteria");
		listAllergyRequest.appendChild(medicationAllergyIntoleranceSearchCriteria);
		
		Element patient = bodyDocument.createElement("patient");
		medicationAllergyIntoleranceSearchCriteria.appendChild(patient);
		
		Element system = bodyDocument.createElement("system");
		system.setTextContent(cprSystemProperty);
		patient.appendChild(system);
		
		Element code = bodyDocument.createElement("code");
		code.setTextContent(cprProperty);
		patient.appendChild(code);
		
		Element requesterOrganization = bodyDocument.createElement("requesterOrganization");
		medicationAllergyIntoleranceSearchCriteria.appendChild(requesterOrganization);
		
		Element orgType = bodyDocument.createElement("type");
		orgType.setAttribute("value", orgTypeProperty);
		requesterOrganization.appendChild(orgType);
		
		Element orgCode = bodyDocument.createElement("code");
		orgCode.setTextContent(orgCodeProperty);
		requesterOrganization.appendChild(orgCode);

		Element consentOverride = bodyDocument.createElement("consentOverride");
		consentOverride.setTextContent(consetOverride);
		medicationAllergyIntoleranceSearchCriteria.appendChild(consentOverride);
		
		return bodyDocument;
	}

	private String getRandomCPR(String cprFileProperty) {
		loadCPRList(cprFileProperty);
		String cpr = cprCache.get(RANDOM.nextInt(cprCache.size()));
		getThreadContext().getVariables().put(getClass().getName() + ".cpr", cpr);
		return cpr;
	}

	private synchronized void loadCPRList(String cprFileProperty) {
		if (cprCache == null) {
			BufferedReader reader = null;
			try {
				InputStream cprInputStream;
				File cprFile = new File(cprFileProperty);
				if (cprFile.exists() && cprFile.canRead()) {
					cprInputStream = new BufferedInputStream(new FileInputStream(cprFile));
				} else {
					cprInputStream = new BufferedInputStream(getClass().getResourceAsStream(cprFileProperty));
				}
				reader = new BufferedReader(new InputStreamReader(cprInputStream));
				Set<String> cpr = new HashSet<String>();
				for (String line = reader.readLine(); line != null; line = reader.readLine()) {
					line = line.trim();
					if (!line.isEmpty()) {
						cpr.add(line);
					}
				}
				ListAllergyRequestSampler.cprCache = new ArrayList<String>(cpr);
			} catch (IOException e) {
				throw new IllegalStateException(e);
			} finally {
				if (reader != null) {
					try {
						reader.close();
					} catch (IOException ignore) {
					}
				}
			}
		}
	}
}
