/*
 * The contents of this file is the CONFIDENTAL and PROPRIETARY
 * property of Arosii Information Systems A/S. Any unauthorized use,
 * reproduction or transfer of this file is strictly prohibited.
 *
 * Copyright (C) 2012 Arosii Information Systems A/S.  All rights reserved.
 * This is an unpublished work, and is subject to limited
 * distribution and restricted disclosure only.
 */
package com.arosii.jmeter.protocol.minlog.sampler;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.jmeter.protocol.http.sampler.HTTPSampleResult;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.arosii.jmeter.protocol.nsp.sampler.AbstractNSPSoapSampler;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.xml.XmlUtil;

/**
 * 
 * @author Jacob
 */
public class ListLogStatementsRequestSampler extends AbstractNSPSoapSampler {

	private static final long serialVersionUID = 1L;

	public static final String CPR = "ListLogStatementsRequestSampler.CPR";

	public static final String CPR_FILE = "ListLogStatementsRequestSampler.CPRFile";

	public static final String ON_BEHALF_OF = "ListLogStatementsRequestSampler.OnBehalfOf";

	public static final String PAGE_SIZE = "ListLogStatementsRequestSampler.PageSize";

	public static final String PAGE_START_AFTER = "ListLogStatementsRequestSampler.PageStartAfter";

	public static final Random RANDOM = new Random();

	private static List<String> cprCache;

	public ListLogStatementsRequestSampler() {
		setPath("/minlog/");
		setSOAPAction("http://nsi.dk/minlog/2014/05/23/ListLogStatements");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected HTTPSampleResult sample(URL url, String method, boolean areFollowingRedirect, int frameDepth) {
		clearReply();
		setXmlData(createRequest());
		HTTPSampleResult sample = super.sample(url, method, areFollowingRedirect, frameDepth);
		putReply(sample);
		return sample;
	}

	private String createRequest() {
		SOSIFactory sosiFactory = getSOSIFactory();
		IDCard idCard = getIDCard();

		// Create a new request and add the id card to it.
		Request request = sosiFactory.createNewRequest(false, null);
		request.setIDCard(idCard);

		Document bodyDocument = createListLogStatementsRequestBody();

		request.setBody(bodyDocument.getDocumentElement());

		Document doc = request.serialize2DOMDocument();

		return XmlUtil.node2String(doc, false, true);
	}

	private Document createListLogStatementsRequestBody() {
		Document bodyDocument;
		try {
			bodyDocument = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
		} catch (ParserConfigurationException e) {
			throw new IllegalStateException(e);
		}
		String requestMethod = getPropertyAsBoolean(ON_BEHALF_OF, false) ? "ListLogStatementsOnBehalfOfRequest" : "ListLogStatementsRequest";
		Element listLogStatementsRequest = bodyDocument.createElementNS("http://nsi.dk/minlog/2014/05/23/", requestMethod);
		listLogStatementsRequest.setPrefix("ml");
		bodyDocument.appendChild(listLogStatementsRequest);

		String cprProperty = getPropertyAsString(CPR);
		String cprFileProperty = getPropertyAsString(CPR_FILE);
		String pageSizeProperty = getPropertyAsString(PAGE_SIZE);
		String pageStartAfterProperty = getPropertyAsString(PAGE_START_AFTER);

		if (cprFileProperty != null && !cprFileProperty.trim().isEmpty()) {
			cprProperty = getRandomCPR(cprFileProperty, pageStartAfterProperty);
		}

		Element civilRegistrationNumber = bodyDocument.createElement("cprNR");
		civilRegistrationNumber.setTextContent(cprProperty);
		listLogStatementsRequest.appendChild(civilRegistrationNumber);

		if (pageSizeProperty != null && !pageSizeProperty.trim().isEmpty()) {
			Element pageSize = bodyDocument.createElement("pageSize");
			pageSize.setTextContent(pageSizeProperty);
			listLogStatementsRequest.appendChild(pageSize);
		}

		if (pageStartAfterProperty != null && !pageStartAfterProperty.trim().isEmpty()) {
			Element pageStartAfter = bodyDocument.createElement("pageStartAfter");
			pageStartAfter.setTextContent(pageStartAfterProperty);
			listLogStatementsRequest.appendChild(pageStartAfter);
		}

		return bodyDocument;
	}

	private String getRandomCPR(String cprFileProperty, String pageStartAfterProperty) {
		loadCPRList(cprFileProperty);
		if (pageStartAfterProperty != null && !pageStartAfterProperty.trim().isEmpty()) {
			String previousCpr = getThreadContext().getVariables().get(getClass().getName() + ".cpr");
			return previousCpr;
		} else {
			String cpr = cprCache.get(RANDOM.nextInt(cprCache.size()));
			getThreadContext().getVariables().put(getClass().getName() + ".cpr", cpr);
			return cpr;
		}
	}

	private synchronized void loadCPRList(String cprFileProperty) {
		if (cprCache == null) {
			BufferedReader reader = null;
			try {
				InputStream cprInputStream;
				File cprFile = new File(cprFileProperty);
				if (cprFile.exists() && cprFile.canRead()) {
					cprInputStream = new BufferedInputStream(new FileInputStream(cprFile));
				} else {
					cprInputStream = new BufferedInputStream(getClass().getResourceAsStream(cprFileProperty));
				}
				reader = new BufferedReader(new InputStreamReader(cprInputStream));
				Set<String> cpr = new HashSet<String>();
				for (String line = reader.readLine(); line != null; line = reader.readLine()) {
					line = line.trim();
					if (!line.isEmpty()) {
						cpr.add(line);
					}
				}
				ListLogStatementsRequestSampler.cprCache = new ArrayList<String>(cpr);
			} catch (IOException e) {
				throw new IllegalStateException(e);
			} finally {
				if (reader != null) {
					try {
						reader.close();
					} catch (IOException ignore) {
					}
				}
			}
		}
	}
}
