/**
* The MIT License
*   
* Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
*   
* Copyright (C) 2018 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
*   
*   
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*   
*   
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*   
*   
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*/

package dk.nsp.jmeter.protocol.nas.sampler;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.jmeter.protocol.http.sampler.HTTPSampleResult;
import org.apache.jorphan.logging.LoggingManager;
import org.apache.log.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.arosii.jmeter.protocol.nsp.sampler.AbstractNSPSoapSampler;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.xml.XmlUtil;

/**
 * 
 * @author
 */
public class PullPointRequestSampler extends AbstractNSPSoapSampler {

	protected static final long serialVersionUID = 1L;

	protected static final Logger log = LoggingManager.getLoggerFor("jmeter.protocol.nas");

	public static final String PULL_POINT = "PullPointRequestSampler.PullPoint";

	public static final String PULL_POINT_AUTO = "PullPointRequestSampler.PullPointAutoGenerate";

	public static final String MAXIMUM = "PullPointRequestSampler.Maximum";

	public static final String PULL_POINT_FILE = "PullPointRequestSampler.PullPointFile";

	public static final String NO_OF_PULL_POINTS = "PullPointRequestSampler.NoOfPullPoints";

	public static final String NO_OF_PULL_POINTS_DB = "PullPointRequestSampler.NoOfPullPointsInDatabase";

	public static final String PREFIX_PULL_POINTS = "PullPointRequestSampler.PrefixPullPoints";

	public static final Random RANDOM = new Random();

	protected static BlockingQueue<Integer> availablePullPointNumbers;

	protected static ThreadLocal<Integer> currentPullPointNumber = new ThreadLocal<>();

	protected static List<String> pullPointCache;

	public PullPointRequestSampler() {
		setPath("/nas2/pullpoint/");
		setSOAPAction("http://http://docs.oasis-open.org/wsn/bw-2/PullPoint/GetMessagesRequest");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String getPath() {
		String pullPointId = getPullPoint();
		String adjustedPath = super.getPath() + pullPointId;
		return adjustedPath;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected HTTPSampleResult sample(URL url, String method, boolean areFollowingRedirect, int frameDepth) {
		try {
			clearReply();
			setXmlData(createRequest());
			HTTPSampleResult sample = super.sample(url, method, areFollowingRedirect, frameDepth);
			putReply(sample);
			return sample;
		} finally {
			if (currentPullPointNumber.get() != null) {
				returnPullPointNumber(currentPullPointNumber.get());
				currentPullPointNumber.set(null);
			}
		}
	}

	protected String createRequest() {
		SOSIFactory sosiFactory = getSOSIFactory();
		IDCard idCard = getIDCard();

		// Create a new request and add the id card to it.
		Request request = sosiFactory.createNewRequest(false, null);
		request.setIDCard(idCard);

		Document bodyDocument = createPullPointRequestBody();

		request.setBody(bodyDocument.getDocumentElement());

		Document doc = request.serialize2DOMDocument();

		return XmlUtil.node2String(doc, false, true);
	}

	protected Document createPullPointRequestBody() {
		Document bodyDocument;
		try {
			bodyDocument = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
		} catch (ParserConfigurationException e) {
			throw new IllegalStateException(e);
		}
		String requestMethod = "GetMessages";
		Element getMethodRequest = bodyDocument.createElementNS("http://docs.oasis-open.org/wsn/b-2", requestMethod);
		getMethodRequest.setPrefix("ns4");
		bodyDocument.appendChild(getMethodRequest);

		String maximumNumberProperty = getPropertyAsString(MAXIMUM);

		Element notificationMessage = bodyDocument.createElement("NotificationMessage");
		getMethodRequest.appendChild(notificationMessage);

		Element maximumNumber = bodyDocument.createElement("ns4:MaximumNumber");
		maximumNumber.setTextContent(maximumNumberProperty);
		notificationMessage.appendChild(maximumNumber);

		return bodyDocument;

	}

	protected String getPullPoint() {
		String pullPointProperty = getPropertyAsString(PULL_POINT);
		String pullPointFileProperty = getPropertyAsString(PULL_POINT_FILE);
		String pullPointAutoProperty = getPropertyAsString(PULL_POINT_AUTO);

		if (pullPointFileProperty != null && !pullPointFileProperty.trim().isEmpty()) {
			return getRandomPullPointFromFile(pullPointFileProperty);
		}
		if (pullPointAutoProperty != null && pullPointAutoProperty.equals("true")) {
			return getRandomPullPoint();
		}

		return pullPointProperty;
	}

	protected String getRandomPullPointFromFile(String pullPointFileProperty) {
		loadPullPointList(pullPointFileProperty);
		String pullPoint = pullPointCache.get(RANDOM.nextInt(pullPointCache.size()));
		getThreadContext().getVariables().put(getClass().getName() + ".pullPoint", pullPoint);
		return pullPoint;
	}

	protected String getRandomPullPoint() {
		String prefixPullPointsProperty = getPropertyAsString(PREFIX_PULL_POINTS);
		String noOfPullPointsInDatabaseProperty = getPropertyAsString(NO_OF_PULL_POINTS_DB);

		int noOfPullPointsInDatabase = Integer.parseInt(noOfPullPointsInDatabaseProperty);

		int pullpointNumber = borrowPullPointNumber();

		int noOfZeroesInFormat = Integer.toString(noOfPullPointsInDatabase).length();
		String randomFormatted = String.format(("%0" + Integer.toString(noOfZeroesInFormat) + "d"), pullpointNumber + 1);

		return prefixPullPointsProperty + randomFormatted;
	}

	protected int borrowPullPointNumber() {
		String noOfPullPointsProperty = getPropertyAsString(NO_OF_PULL_POINTS);
		int noOfPullPoints = Integer.parseInt(noOfPullPointsProperty);
		int nodeId = getNodeId();
		int nodeCount = getNodeCount();

		int pullPointNumber = borrowPullPointNumber(noOfPullPoints, nodeId, nodeCount);
		currentPullPointNumber.set(pullPointNumber);
		return pullPointNumber;
	}

	protected static int borrowPullPointNumber(int noOfPullPoints, int nodeId, int nodeCount) {
		initPullPointNumberQueue(noOfPullPoints, nodeId, nodeCount);
		try {
			Integer pullPointNumber = availablePullPointNumbers.poll(1, TimeUnit.MINUTES);
			if (pullPointNumber == null) {
				throw new IllegalStateException("No Pullpoint number became available.");
			}
			return pullPointNumber;
		} catch (InterruptedException ignored) {
			throw new IllegalStateException("Interrupted waiting for pullpoint number to become available.");
		}
	}

	protected static void returnPullPointNumber(int pullPointNumber) {
		availablePullPointNumbers.add(pullPointNumber);
	}

	protected static synchronized void initPullPointNumberQueue(int noOfPullPoints, int nodeId, int nodeCount) {
		if (availablePullPointNumbers == null) {
			int pullPointsPerNode = (noOfPullPoints / nodeCount) + 1;
			List<Integer> pullPointNumbers = new ArrayList<>(pullPointsPerNode);
			for (int i = 0; i < pullPointsPerNode; i++) {
				int pullpointNumber = (i * nodeCount) + nodeId;
				if (pullpointNumber < noOfPullPoints) {
					pullPointNumbers.add(pullpointNumber);
				}
			}
			
			StringBuilder sb = new StringBuilder();
			for (int pullpointNumber : pullPointNumbers) {
				sb.append(pullpointNumber).append(", ");
			}
			log.info("Available pullpoint numbers: " + sb.substring(0, sb.length() - 2));
			
			Collections.shuffle(pullPointNumbers);
			availablePullPointNumbers = new LinkedBlockingQueue<>(pullPointNumbers);
		}
	}

	protected synchronized void loadPullPointList(String pullPointFileProperty) {
		if (pullPointCache == null) {
			BufferedReader reader = null;
			try {
				InputStream pullPointInputStream;
				File pullPointFile = new File(pullPointFileProperty);
				if (pullPointFile.exists() && pullPointFile.canRead()) {
					pullPointInputStream = new BufferedInputStream(new FileInputStream(pullPointFile));
				} else {
					pullPointInputStream = new BufferedInputStream(getClass().getResourceAsStream(pullPointFileProperty));
				}
				reader = new BufferedReader(new InputStreamReader(pullPointInputStream));
				Set<String> pullPoint = new HashSet<String>();
				for (String line = reader.readLine(); line != null; line = reader.readLine()) {
					line = line.trim();
					if (!line.isEmpty()) {
						pullPoint.add(line);
					}
				}
				PullPointRequestSampler.pullPointCache = new ArrayList<String>(pullPoint);
			} catch (IOException e) {
				throw new IllegalStateException(e);
			} finally {
				if (reader != null) {
					try {
						reader.close();
					} catch (IOException ignore) {
					}
				}
			}
		}
	}

}