/*
 * The contents of this file is the CONFIDENTAL and PROPRIETARY
 * property of Arosii Information Systems A/S. Any unauthorized use,
 * reproduction or transfer of this file is strictly prohibited.
 *
 * Copyright (C) 2012 Arosii Information Systems A/S.  All rights reserved.
 * This is an unpublished work, and is subject to limited
 * distribution and restricted disclosure only.
 */
package com.arosii.jmeter.protocol.sdm.extractor.registers;

import java.util.List;

import org.apache.jmeter.testelement.TestElement;
import org.apache.jmeter.threads.JMeterContext;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import com.arosii.jmeter.protocol.sdm.extractor.ReplicationRequestExtractor;
import com.arosii.jmeter.protocol.sdm.sampler.ReplicationRequestSampler;

public class AuthorizationRegister {

	public static class Authorization {

		private final String cpr;

		private final String firstName;

		private final String lastName;

		public Authorization(String cpr, String firstName, String lastName) {
			this.cpr = cpr;
			this.firstName = firstName;
			this.lastName = lastName;
		}

		public String getCpr() {
			return cpr;
		}

		public String getFirstName() {
			return firstName;
		}

		public String getLastName() {
			return lastName;
		}

		public static enum Field {

			CPR("cpr"), FIRST_NAME("fornavn"), LAST_NAME("efternavn");

			private final String tag;

			private Field(String tag) {
				this.tag = tag;
			}

			public String property() {
				return "ReplicationRequestExtractor.Authorization." + name();
			}
		}

		public static class Parser extends Extractor<Authorization> {

			private final boolean extractCpr;

			private final boolean extractFirstName;

			private final boolean extractLastName;

			private final StringBuilder cpr = new StringBuilder();

			private final StringBuilder firstName = new StringBuilder();

			private final StringBuilder lastName = new StringBuilder();

			private Field state = null;

			public Parser(String registerVariable, String extractionNotCompleteVariable, String revisionNumberVariable, boolean extractCpr, boolean extractFirstName, boolean extractLastName, int extractionPercentage, int extractionLimit, int readLimit, int expectedResponseSize) {
				super(registerVariable, extractionNotCompleteVariable, revisionNumberVariable, extractionPercentage, extractionLimit, readLimit, expectedResponseSize);
				this.extractCpr = extractCpr;
				this.extractFirstName = extractFirstName;
				this.extractLastName = extractLastName;
			}

			/**
			 * {@inheritDoc}
			 */
			@Override
			public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
				super.startElement(uri, localName, qName, attributes);
				state = null;
				if (localName.equals("autorisation")) {
					cpr.setLength(0);
					firstName.setLength(0);
					lastName.setLength(0);
				} else if (extractCpr && localName.equals(Field.CPR.tag)) {
					state = Field.CPR;
				} else if (extractFirstName && localName.equals(Field.FIRST_NAME.tag)) {
					state = Field.FIRST_NAME;
				} else if (extractLastName && localName.equals(Field.LAST_NAME.tag)) {
					state = Field.LAST_NAME;
				}
			}

			/**
			 * {@inheritDoc}
			 */
			@Override
			public void characters(char[] ch, int start, int length) throws SAXException {
				super.characters(ch, start, length);
				if (state == Field.CPR) {
					cpr.append(ch, start, length);
				} else if (state == Field.FIRST_NAME) {
					firstName.append(ch, start, length);
				} else if (state == Field.LAST_NAME) {
					lastName.append(ch, start, length);
				}
			}

			/**
			 * {@inheritDoc}
			 */
			@Override
			public void endElement(String uri, String localName, String qName) throws SAXException {
				super.endElement(uri, localName, qName);
				if (localName.equals("autorisation")) {
					if (cpr.length() > 0 || firstName.length() > 0 || lastName.length() > 0) {
						addEntry(new Authorization(cpr.toString(), firstName.toString(), lastName.toString()));
					}
				}
			}
		}

		/**
		 * @param replicationRequestExtractor
		 */
		public static Extractor<Authorization> createParser(TestElement element) {
			String registerVariable = element.getPropertyAsString(ReplicationRequestExtractor.REGISTER_VARIABLE, "");
			String extractionNotCompleteVariable = element.getPropertyAsString(ReplicationRequestExtractor.EXTRACTION_NOT_COMPLETE_VARIABLE, "");
			String revisionNumberVariable = element.getPropertyAsString(ReplicationRequestExtractor.REVISION_NUMBER_VARIABLE, "");
			int extractionPercentage = 100;
			try {
				extractionPercentage = Integer.valueOf(element.getPropertyAsString(ReplicationRequestExtractor.EXTRACTED_PERCENTAGE, "100"));
			} catch (NumberFormatException ignore) {
			}
			int extractionLimit = Integer.MAX_VALUE;
			try {
				extractionLimit = Integer.valueOf(element.getPropertyAsString(ReplicationRequestExtractor.MAX_EXTRACTED, String.valueOf(Integer.MAX_VALUE)));
			} catch (NumberFormatException ignore) {
			}
			int readLimit = Integer.MAX_VALUE;
			try {
				readLimit = Integer.valueOf(element.getPropertyAsString(ReplicationRequestExtractor.MAX_READ, String.valueOf(Integer.MAX_VALUE)));
			} catch (NumberFormatException ignore) {
			}

			boolean extractCpr = element.getPropertyAsBoolean(Field.CPR.property(), false);
			boolean extractFirstName = element.getPropertyAsBoolean(Field.FIRST_NAME.property(), false);
			boolean extractLastName = element.getPropertyAsBoolean(Field.LAST_NAME.property(), false);

			String expectedResponseSizeProperty = element.getPropertyAsString(ReplicationRequestSampler.MAX_ROW_COUNT, "");
			int expectedResponseSize;
			if (expectedResponseSizeProperty.length() > 0) {
				expectedResponseSize = Integer.valueOf(expectedResponseSizeProperty);
			} else {
				expectedResponseSize = Integer.MAX_VALUE;
			}

			if ((registerVariable.length() == 0) && extractionNotCompleteVariable.length() == 0 && revisionNumberVariable.length() == 0) {
				return null; // Nothing to do
			} else {
				return new Parser(registerVariable, extractionNotCompleteVariable, revisionNumberVariable, extractCpr, extractFirstName, extractLastName, extractionPercentage, extractionLimit, readLimit, expectedResponseSize);
			}
		}

		@SuppressWarnings("unchecked")
		public static Authorization getRandomAuthorization(JMeterContext context, String variable) {
			List<List<Authorization>> items = (List<List<Authorization>>) context.getVariables().getObject(variable);
			if (items != null && !items.isEmpty()) {
				int escape = 0;
				List<Authorization> randomAuthorizations;
				do {
					randomAuthorizations = items.get(Extractor.RANDOM.nextInt(items.size()));
				} while (randomAuthorizations.isEmpty() && escape++ < items.size() * 2);
				if (!randomAuthorizations.isEmpty()) {
					return randomAuthorizations.get(Extractor.RANDOM.nextInt(randomAuthorizations.size()));
				}
			}
			return null;
		}
	}
}