/*
 * The contents of this file is the CONFIDENTAL and PROPRIETARY
 * property of Arosii Information Systems A/S. Any unauthorized use,
 * reproduction or transfer of this file is strictly prohibited.
 *
 * Copyright (C) 2012 Arosii Information Systems A/S.  All rights reserved.
 * This is an unpublished work, and is subject to limited
 * distribution and restricted disclosure only.
 */
package com.arosii.jmeter.protocol.sdm.extractor.registers;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Random;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.jmeter.threads.JMeterVariables;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * 
 * @author Jacob
 */
public abstract class Extractor<E> extends DefaultHandler implements ContentHandler {

	public static final Random RANDOM = new Random();

	private static final Pattern ATOM_DATE_PATTERN = Pattern.compile("(\\d{4})(?:-(\\d{2}))?(?:-(\\d{2}))?(?:([Tt])?(?:(\\d{2}))?(?::(\\d{2}))?(?::(\\d{2}))?(?:\\.(\\d{3}))?)?([Zz])?(?:([+-])(\\d{2}):(\\d{2}))?");

	private static final String ATOM_NAMESPACE = "http://www.w3.org/2005/Atom";

	private final String registerVariable;

	private final String registerSizeVariable;

	private final String extractionNotCompleteVariable;

	private final String readCountVariable;

	private final String revisionNumberVariable;

	private final int extractionLimit;

	private final int readLimit;

	private final int expectedResponseSize;

	private final StringBuilder feedId = new StringBuilder();

	private final StringBuilder entryId = new StringBuilder();

	private final StringBuilder validFrom = new StringBuilder();

	private final StringBuilder validTo = new StringBuilder();

	private AtomState atomState;

	private int responseSize = 0;

	protected final int extractionPercentage;

	protected final Date now;

	protected final List<E> entries = new ArrayList<E>();

	private static enum AtomState {
		FEED, ENTRY, FEED_ID, ENTRY_ID, CONTENT, CONTENT_VALID_FROM, CONTENT_VALID_TO;
	}

	public Extractor(String registerVariable, String extractionNotCompleteVariable, String revisionNumberVariable, int extractionPercentage, int extractionLimit, int readLimit, int expectedResponseSize) {
		this.registerVariable = registerVariable;
		this.registerSizeVariable = "_" + registerVariable + "_SIZE_";
		this.extractionNotCompleteVariable = extractionNotCompleteVariable;
		this.readCountVariable = "_" + extractionNotCompleteVariable + "_COUNT_";
		this.revisionNumberVariable = revisionNumberVariable;
		this.extractionPercentage = extractionPercentage;
		this.extractionLimit = extractionLimit;
		this.readLimit = readLimit;
		this.expectedResponseSize = expectedResponseSize;
		this.now = new Date();
	}

	@SuppressWarnings("unchecked")
	public void saveVariables(JMeterVariables variables) {
		boolean extractionLimitReached = false;
		if (registerVariable != null && registerVariable.length() > 0) {
			List<List<E>> allEntries = (List<List<E>>) variables.getObject(registerVariable);
			Integer size = (Integer) variables.getObject(registerSizeVariable);
			if (allEntries == null) {
				allEntries = new ArrayList<List<E>>();
				size = new Integer(0);
				variables.putObject(registerVariable, allEntries);
				variables.putObject(registerSizeVariable, size);
			}
			if (size + entries.size() < extractionLimit) {
				allEntries.add(entries);
				variables.putObject(registerSizeVariable, size + entries.size());
			} else {
				extractionLimitReached = true;
			}
		}
		if (extractionNotCompleteVariable != null && extractionNotCompleteVariable.length() > 0) {
			Integer readCount = (Integer) variables.getObject(readCountVariable);
			if (readCount == null) {
				readCount = 0;
			}
			readCount = readCount + responseSize;
			variables.putObject(readCountVariable, readCount);
			boolean extractionNotComplete = responseSize == expectedResponseSize && !extractionLimitReached && readCount < readLimit;
			variables.put(extractionNotCompleteVariable, String.valueOf(extractionNotComplete));
		}
		if (revisionNumberVariable != null && revisionNumberVariable.length() > 0) {
			// Invariant: entryId contains the last read <atom:entry><atom:id>XXXXXXXXXX</atom:id></atom:entry> value
			if (entryId.length() == 0 || feedId.length() == 0) {
				variables.remove(revisionNumberVariable);
			} else {
				// A '/' is added before the revision
				String revision = entryId.toString().substring(feedId.toString().length() + 1);
				variables.put(revisionNumberVariable, revision);
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
		if (localName.equals("feed") && uri.equals(ATOM_NAMESPACE)) {
			atomState = AtomState.FEED;
			feedId.setLength(0);
		} else if (localName.equals("entry") && uri.equals(ATOM_NAMESPACE)) {
			atomState = AtomState.ENTRY;
			entryId.setLength(0);
		} else if (localName.equals("id") && uri.equals(ATOM_NAMESPACE)) {
			if (atomState == AtomState.FEED) {
				atomState = AtomState.FEED_ID;
			} else if (atomState == AtomState.ENTRY) {
				atomState = AtomState.ENTRY_ID;
			}
		} else if (localName.equals("content") && uri.equals(ATOM_NAMESPACE)) {
			atomState = AtomState.CONTENT;
			validFrom.setLength(0);
			validTo.setLength(0);
		} else if (localName.equals("validFrom") && atomState == AtomState.CONTENT) {
			atomState = AtomState.CONTENT_VALID_FROM;
		} else if (localName.equals("validTo") && atomState == AtomState.CONTENT) {
			atomState = AtomState.CONTENT_VALID_TO;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void characters(char[] ch, int start, int length) throws SAXException {
		if (atomState == AtomState.FEED_ID) {
			feedId.append(ch, start, length);
		} else if (atomState == AtomState.ENTRY_ID) {
			entryId.append(ch, start, length);
		} else if (atomState == AtomState.CONTENT_VALID_FROM) {
			validFrom.append(ch, start, length);
		} else if (atomState == AtomState.CONTENT_VALID_TO) {
			validTo.append(ch, start, length);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void endElement(String uri, String localName, String qName) throws SAXException {
		if (localName.equals("feed") && uri.equals(ATOM_NAMESPACE)) {
			atomState = null;
		} else if (localName.equals("entry") && uri.equals(ATOM_NAMESPACE)) {
			atomState = AtomState.FEED;
		} else if (localName.equals("id") && uri.equals(ATOM_NAMESPACE)) {
			if (atomState == AtomState.FEED_ID) {
				atomState = AtomState.FEED;
			} else if (atomState == AtomState.ENTRY_ID) {
				atomState = AtomState.ENTRY;
			}
		} else if (localName.equals("content") && uri.equals(ATOM_NAMESPACE)) {
			responseSize++;
			atomState = AtomState.ENTRY;
		} else if (localName.equals("validFrom") && atomState == AtomState.CONTENT_VALID_FROM) {
			atomState = AtomState.CONTENT;
		} else if (localName.equals("validTo") && atomState == AtomState.CONTENT_VALID_TO) {
			atomState = AtomState.CONTENT;
		}
	}

	protected boolean addEntry(E entry) {
		if (validFrom.length() > 0 && validTo.length() > 0) {
			if (parse(validFrom.toString()).before(now) && parse(validTo.toString()).after(now)) {
				if (RANDOM.nextInt(100) < extractionPercentage) {
					return entries.add(entry);
				}
			}
		}
		return false;
	}

	/**
	 * Parse the serialized string form into a java.util.Date
	 * <p>
	 * Code from org.apache.abdera.model.AtomDate
	 */
	protected Date parse(String date) {
		Matcher m = ATOM_DATE_PATTERN.matcher(date);
		if (m.find()) {
			if (m.group(4) == null) {
				throw new IllegalArgumentException("Invalid Date Format");
			}
			Calendar c = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
			int hoff = 0, moff = 0, doff = -1;
			if (m.group(10) != null) {
				doff = m.group(10).equals("-") ? 1 : -1;
				hoff = doff * (m.group(11) != null ? Integer.parseInt(m.group(11)) : 0);
				moff = doff * (m.group(12) != null ? Integer.parseInt(m.group(12)) : 0);
			}
			c.set(Calendar.YEAR, Integer.parseInt(m.group(1)));
			c.set(Calendar.MONTH, m.group(2) != null ? Integer.parseInt(m.group(2)) - 1 : 0);
			c.set(Calendar.DATE, m.group(3) != null ? Integer.parseInt(m.group(3)) : 1);
			c.set(Calendar.HOUR_OF_DAY, m.group(5) != null ? Integer.parseInt(m.group(5)) + hoff : 0);
			c.set(Calendar.MINUTE, m.group(6) != null ? Integer.parseInt(m.group(6)) + moff : 0);
			c.set(Calendar.SECOND, m.group(7) != null ? Integer.parseInt(m.group(7)) : 0);
			c.set(Calendar.MILLISECOND, m.group(8) != null ? Integer.parseInt(m.group(8)) : 0);
			return c.getTime();
		} else {
			throw new IllegalArgumentException("Invalid Date Format");
		}
	}

}
