package com.arosii.jmeter.protocol.sor.sampler;

import com.arosii.jmeter.protocol.nsp.sampler.AbstractNSPSoapSampler;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.Reply;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.jmeter.protocol.http.sampler.HTTPSampleResult;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public abstract class AbstractSORSampler extends AbstractNSPSoapSampler {

    protected AbstractSORSampler(String path, String soapAction) {
        setPath(path);
        setSendSOAPAction(true);
        setSOAPAction(soapAction);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected HTTPSampleResult sample(java.net.URL url, String method, boolean areFollowingRedirect, int frameDepth) {
        clearReply();
        setXmlData(createRequest());
        HTTPSampleResult sample = super.sample(url, method, areFollowingRedirect, frameDepth);
        putReply(sample);
        validateReply(sample);
        return sample;
    }

    protected void validateReply(HTTPSampleResult sample) {
        Reply reply = getSOSIFactory().deserializeReply(sample.getResponseDataAsString());

        if(reply.isFault()) {
            sample.setSuccessful(false);
            sample.setResponseMessage(reply.getFaultCode() + " / " + reply.getFaultString());
        } else {
            validateStatus(reply, sample);

            if(sample.isSuccessful()) {
                validateContent(reply, sample);
            }
        }
    }

    protected void validateStatus(Reply reply, HTTPSampleResult sample) {
        NodeList statusNodes = reply.getBody().getElementsByTagNameNS("http://sundhedsdatastyrelsen.dk/SORLookupService/2018/09/06/", "Status");
        if (statusNodes != null && statusNodes.getLength() > 0) {
            for (int i = 0; i < statusNodes.getLength(); i++) {
                if (statusNodes.item(i).getNodeType() == Node.ELEMENT_NODE) {
                    Element statusElement = (Element) statusNodes.item(i);
                    NodeList errorCodeNodes = statusElement.getElementsByTagNameNS("http://sundhedsdatastyrelsen.dk/SORLookupService/2018/09/06/", "ErrorCode");
                    NodeList messageNodes = statusElement.getElementsByTagNameNS("http://sundhedsdatastyrelsen.dk/SORLookupService/2018/09/06/", "Message");
                    if(errorCodeNodes != null && errorCodeNodes.getLength() > 0) {
                        for (int j = 0; j < errorCodeNodes.getLength(); j++) {
                            if (errorCodeNodes.item(j).getNodeType() == Node.ELEMENT_NODE) {
                                Element errorCodeElement = (Element) errorCodeNodes.item(i);
                                String errorCode = errorCodeElement.getTextContent();
                                if (errorCode != null && !errorCode.trim().equals("0")) {
                                    StringBuilder messageBuilder = new StringBuilder();
                                    for (int k = 0; k < messageNodes.getLength(); k++) {
                                        if (messageNodes.item(k).getNodeType() == Node.ELEMENT_NODE) {
                                            Element messageNode = (Element) messageNodes.item(k);
                                            String message = messageNode.getTextContent();
                                            messageBuilder.append(message).append(" / ");
                                        }
                                    }
                                    sample.setSuccessful(false);
                                    sample.setResponseMessage("Status was not successful: " + errorCode + " - " + messageBuilder.toString());
                                }
                            } else {
                                sample.setSuccessful(false);
                                sample.setResponseMessage("ErrorCode node in response was not an element node");
                            }
                        }
                    } else {
                        sample.setSuccessful(false);
                        sample.setResponseMessage("Response missing ErrorCode node in Status node");
                    }
                } else {
                    sample.setSuccessful(false);
                    sample.setResponseMessage("Status node in response was not an element node");
                }
            }
        } else {
            sample.setSuccessful(false);
            sample.setResponseMessage("Response missing Status node");
        }
    }

    protected abstract void validateContent(Reply reply, HTTPSampleResult sample);

    private String createRequest() {
        SOSIFactory sosiFactory = getSOSIFactory();
        IDCard idCard = getIDCard();

        // Create a new request and add the id card to it.
        Request request = sosiFactory.createNewRequest(false, null);
        request.setIDCard(idCard);

        Document bodyDocument = createRequestBody();

        request.setBody(bodyDocument.getDocumentElement());

        Document doc = request.serialize2DOMDocument();

        return XmlUtil.node2String(doc, false, true);
    }

    protected abstract Document createRequestBody();
}
