/*
 * The contents of this file is the CONFIDENTAL and PROPRIETARY
 * property of Arosii Information Systems A/S. Any unauthorized use,
 * reproduction or transfer of this file is strictly prohibited.
 *
 * Copyright (C) 2012 Arosii Information Systems A/S.  All rights reserved.
 * This is an unpublished work, and is subject to limited
 * distribution and restricted disclosure only.
 */
package com.arosii.jmeter.protocol.sts.sampler;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.Date;
import java.util.Properties;

import org.apache.jmeter.protocol.http.sampler.HTTPSampleResult;
import org.w3c.dom.Document;

import com.arosii.jmeter.protocol.sts.control.STSIdentity.Type;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.SecurityTokenRequest;
import dk.sosi.seal.model.SignatureConfiguration;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.model.UserInfo;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;

/**
 * 
 * @author Jacob
 */
public class SecurityTokenRequestSampler extends AbstractSTSSoapSampler {

	private static final long serialVersionUID = 1L;

	public static final String CERTIFICATE_PATH = "SecurityTokenRequestSampler.Certificate.Path";

	public static final String CERTIFICATE_PASSWORD = "SecurityTokenRequestSampler.Certificate.Password";

	public static final String CERTIFICATE_TYPE = "SecurityTokenRequestSampler.Certificate.Type";

	public static final String USER_CPR = "SecurityTokenRequestSampler.User.Cpr";

	public static final String USER_GIVENNAME = "SecurityTokenRequestSampler.User.GivenName";

	public static final String USER_SURNAME = "SecurityTokenRequestSampler.User.SurName";

	public static final String USER_EMAIL = "SecurityTokenRequestSampler.User.EMail";

	public static final String USER_OCCUPATION = "SecurityTokenRequestSampler.User.Occupation";

	public static final String USER_ROLE = "SecurityTokenRequestSampler.User.Role";

	public static final String USER_AUTHORIZATIONCODE = "SecurityTokenRequestSampler.User.AuthorizationCode";

	public static final String ORG_TYPE = "SecurityTokenRequestSampler.Org.Type";

	public static final String ORG_ID = "SecurityTokenRequestSampler.Org.Id";

	public static final String ORG_NAME = "SecurityTokenRequestSampler.Org.Name";

	public static final String VERSION = "SecurityTokenRequestSampler.Version";

	public static final String AUTHENTICATION_LEVEL = "SecurityTokenRequestSampler.AuthenticationLevel";

	public static final String VALIDITY_RANGE = "SecurityTokenRequestSampler.ValidityRange";

	public static final String VALIDITY_OFFSET = "SecurityTokenRequestSampler.ValidityOffset";

	public SecurityTokenRequestSampler() {
		setPath("/sts/services/NewSecurityTokenService");
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	protected HTTPSampleResult sample(URL url, String method, boolean areFollowingRedirect, int frameDepth) {
		clearIDCard();
		CredentialVault vault = createCredentialVault();
		SOSIFactory sosiFactory = createSOSIFactory(vault);
		String securityTokenRequest = createSecurityTokenRequest(vault, sosiFactory);
		setXmlData(securityTokenRequest);
		HTTPSampleResult sample = super.sample(url, method, areFollowingRedirect, frameDepth);
		putSOSIFactory(sosiFactory);
		putIDCard(sample);
		return sample;
	}
	
	private CredentialVault createCredentialVault() {
		CredentialVault vault = getVaultFromFile(getPropertyAsString(CERTIFICATE_PATH), getPropertyAsString(CERTIFICATE_PASSWORD));
		putCredentialVault(vault);
		return vault;
	}

	private SOSIFactory createSOSIFactory(CredentialVault vault) {
		Properties props = new Properties(System.getProperties());
		props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, Boolean.toString(false));
		SOSIFactory sosiFactory = new SOSIFactory(new SOSITestFederation(new Properties()), vault, props);
		return sosiFactory;
	}

	private String createSecurityTokenRequest(CredentialVault vault, SOSIFactory sosiFactory) {
		IDCard idCard;
		Type certificateType = Type.valueOf(getPropertyAsString(CERTIFICATE_TYPE));
		switch (certificateType) {
		case MOCES: {
			UserInfo userInfo = new UserInfo(getPropertyAsString(USER_CPR), getPropertyAsString(USER_GIVENNAME), getPropertyAsString(USER_SURNAME), getPropertyAsString(USER_EMAIL), getPropertyAsString(USER_OCCUPATION), getPropertyAsString(USER_ROLE), getPropertyAsString(USER_AUTHORIZATIONCODE));
			CareProvider careProvider = new CareProvider(getPropertyAsString(ORG_TYPE), getPropertyAsString(ORG_ID), getPropertyAsString(ORG_NAME));
			idCard = sosiFactory.createNewUserIDCard("SOSITEST", userInfo, careProvider, AuthenticationLevel.MOCES_TRUSTED_USER, null, null, null, null);
			break;
		}
		case VOCES: {
			CareProvider careProvider = new CareProvider(getPropertyAsString(ORG_TYPE), getPropertyAsString(ORG_ID), getPropertyAsString(ORG_NAME));
			idCard = sosiFactory.createNewSystemIDCard("SOSITEST", careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, null);
			break;
		}
		case FOCES: {
			CareProvider careProvider = new CareProvider(getPropertyAsString(ORG_TYPE), getPropertyAsString(ORG_ID), getPropertyAsString(ORG_NAME));
			idCard = sosiFactory.createNewSystemIDCard("SOSITEST", careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, null);
			break;
		}
		default:
			throw new IllegalArgumentException(certificateType + " not supported yet!");
		}

		// Change the version member of the Seal id card to simulate strange systems
		String version = getPropertyAsString(VERSION, "");
		if (!version.equals("")) {
			setPrivateIDCardVersionMember(idCard, version);
		}

		// Change the authenticationLevel member of the Seal id card to simulate strange systems
		String authenticationLevel = getPropertyAsString(AUTHENTICATION_LEVEL, "");
		if (!authenticationLevel.equals("")) {
			setPrivateIDCardAuthenticationLevelMember(idCard, Integer.valueOf(authenticationLevel));
		}

		// Change the createdDate member of the Seal id card to simulate strange systems
		String validityOffset = getPropertyAsString(VALIDITY_OFFSET, "");
		if (!validityOffset.equals("")) {
			setPrivateIDCardCreatedDateMember(idCard, Integer.valueOf(validityOffset));
		}

		// Change the expiryDate member of the Seal id card to simulate strange systems
		String validityRange = getPropertyAsString(VALIDITY_RANGE, "");
		if (!validityRange.equals("")) {
			setPrivateIDCardExpiryDateMember(idCard, Integer.valueOf(validityRange));
		}
		
		// Create a new request and add the id card to it.
		SecurityTokenRequest request = sosiFactory.createNewSecurityTokenRequest();
		request.setIDCard(idCard);

		// Signing is done in this serialization process when its a Voces certificate (AuthenticationLevel.VOCES_TRUSTED_SYSTEM)
		Document doc = request.serialize2DOMDocument();

		// Otherwise we sign the request manually
		if (certificateType.equals(Type.MOCES) || (certificateType.equals(Type.VOCES) && !authenticationLevel.equals("") && !Integer.valueOf(authenticationLevel).equals(AuthenticationLevel.VOCES_TRUSTED_SYSTEM.getLevel()))) {
			SignatureConfiguration signatureConfiguration = new SignatureConfiguration(new String[] { IDValues.IDCARD }, IDValues.IDCARD, IDValues.id);
			SignatureUtil.sign(SignatureProviderFactory.fromCredentialVault(vault), doc, signatureConfiguration);
		}

		return XmlUtil.node2String(doc, false, true);
	}

	private static void setPrivateIDCardVersionMember(IDCard idCard, String version) {
		try {
			Field versionField = IDCard.class.getDeclaredField("version");
			versionField.setAccessible(true);
			versionField.set(idCard, version);
		} catch (SecurityException e) {
			throw new RuntimeException(e);
		} catch (NoSuchFieldException e) {
			throw new RuntimeException(e);
		} catch (IllegalArgumentException e) {
			throw new RuntimeException(e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}

	private static void setPrivateIDCardAuthenticationLevelMember(IDCard idCard, int level) {
		try {
			Constructor<AuthenticationLevel> constructor = AuthenticationLevel.class.getDeclaredConstructor(Integer.TYPE);
			constructor.setAccessible(true);
			AuthenticationLevel authenticationLevel = constructor.newInstance(level);
			Field authenticationLevelField = IDCard.class.getDeclaredField("authenticationLevel");
			authenticationLevelField.setAccessible(true);
			authenticationLevelField.set(idCard, authenticationLevel);
		} catch (SecurityException e) {
			throw new RuntimeException(e);
		} catch (NoSuchMethodException e) {
			throw new RuntimeException(e);
		} catch (InstantiationException e) {
			throw new RuntimeException(e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		} catch (InvocationTargetException e) {
			throw new RuntimeException(e);
		} catch (NoSuchFieldException e) {
			throw new RuntimeException(e);
		}
	}

	private void setPrivateIDCardCreatedDateMember(IDCard idCard, Integer offset) {
		try {
			Field createdDateField = IDCard.class.getDeclaredField("createdDate");
			createdDateField.setAccessible(true);
			createdDateField.set(idCard, new Date(idCard.getCreatedDate().getTime() + (offset * 60 * 60 * 1000)));
			Field expiryDateField = IDCard.class.getDeclaredField("expiryDate");
			expiryDateField.setAccessible(true);
			expiryDateField.set(idCard, new Date(idCard.getExpiryDate().getTime() + (offset * 60 * 60 * 1000)));
		} catch (SecurityException e) {
			throw new RuntimeException(e);
		} catch (NoSuchFieldException e) {
			throw new RuntimeException(e);
		} catch (IllegalArgumentException e) {
			throw new RuntimeException(e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}

	private static void setPrivateIDCardExpiryDateMember(IDCard idCard, int hours) {
		try {
			Field expiryDateField = IDCard.class.getDeclaredField("expiryDate");
			expiryDateField.setAccessible(true);
			expiryDateField.set(idCard, new Date(idCard.getCreatedDate().getTime() + (hours * 60 * 60 * 1000)));
		} catch (SecurityException e) {
			throw new RuntimeException(e);
		} catch (NoSuchFieldException e) {
			throw new RuntimeException(e);
		} catch (IllegalArgumentException e) {
			throw new RuntimeException(e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}

}
