#!/bin/bash

# Source the profile script which should contain at least a JAVA_HOME
if [ -f ~/.nsp_perf_profile ]; then
	. ~/.nsp_perf_profile
fi

# Ensure JAVA_HOME is set
if [ -z "$JAVA_HOME" ]; then
	echo "Error: JAVA_HOME environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export JAVA_HOME

# Ensure MAVEN_HOME is set
if [ -z "$MAVEN_HOME" ]; then
	echo "Error: MAVEN_HOME environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export MAVEN_HOME

# Ensure M3 is set
if [ -z "$M3" ]; then
	echo "Error: M3 environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export M3

# Ensure SVN_HOME is set
if [ -z "$SVN_HOME" ]; then
	echo "Error: SVN_HOME environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export SVN_HOME

SCRIPT_NAME=`basename $0`

# for checking errors - and exiting if any - call as
#  check_errs $? 'Some error message'
check_errs()
{
  # Function. Parameter 1 is the return code
  # Para. 2 is text to display on failure.
  if [ "$1" -ne "0" ]; then
	echo "ERROR # ${1} : ${2}"
	exit ${1}
  fi
}

# Print usage
usage()
{
  # Function.
	echo "usage: ${SCRIPT_NAME} [-h <hosts configuration file> -p <Port number>] <NSP Component> <Testplan> <Distribution> "
}

# Number of threads in the incremental loops
THREADS=5
MAX_THREADS=30
INC_THREADS=3

# Number of nodes in the incremental loops
NODES=1
MAX_NODES=9
INC_NODES=1

# Number of additional iterations after max tp is found
ADD_ITERATIONS=0

# Read optional arguments using getopts
HOSTS=hosts.properties
PORT=9012
while getopts “h:p:t:n:i:m:o:a:” OPTION; do
	case $OPTION in
		h)
			HOSTS=$OPTARG
			;;
		p)
			PORT=$OPTARG
			;;
		t)
			THREADS=$OPTARG
			;;
		n)
			NODES=$OPTARG
			;;
		i)
			INC_THREADS=$OPTARG
			;;
		m)
			MAX_THREADS=$OPTARG
			;;
		o)
			MAX_NODES=$OPTARG
			;;
		a)
			ADD_ITERATIONS=$OPTARG
			;;
		?)
			usage
			exit
			;;
	esac
done

# Read mandatory arguments using shift and $1
shift $(($OPTIND - 1))
COMPONENT=$1
shift
TESTPLAN=$1
shift
DISTRIBUTION=$1

# Ensure mandatory arguments are provided
if [[ -z $COMPONENT ]] || [[ -z $DISTRIBUTION ]] || [[ -z $TESTPLAN ]]; then
	usage
	exit 2
fi

# A hassle to get current working dir
unamestr=`uname -s`
case "${unamestr}" in
	*Linux*)
		CWD=$(dirname $(readlink -f $0))
	;;
	*Darwin*)
		# readlink works different on OS X so we use a python script instead
		REALPATH=$(python -c 'import os,sys;print os.path.realpath(sys.argv[1])' $0)
		CWD=$(dirname $REALPATH)
	;;
	*CYGWIN*)
		# translate to full cygpath
		CWD=$(cygpath -m $(dirname $(readlink -f $(cygpath $0))))
	;;
	*)
		echo "Unable to detect OS from '${unamestr}'"
		exit 1
	;;
esac
pushd ${CWD} > /dev/null

# Verify input
if [[ ! -d ${CWD}/tests/${COMPONENT} ]]; then
	check_errs 3 "Component '${COMPONENT}' not found."
fi
if [[ ! -r ${CWD}/tests/${COMPONENT}/src/test/jmeter/templates/testplans/${TESTPLAN}.template.jmx ]]; then
	check_errs 3 "Testplan '${TESTPLAN}' not found."
fi
if [[ ! -r ${CWD}/tests/${COMPONENT}/src/test/jmeter/templates/distributions/${DISTRIBUTION}.template.jmx ]]; then
	check_errs 3 "Distribution '${DISTRIBUTION}' not found."
fi

# Log file
STARTED=`date +%Y%m%d_%H%M%S`
LOG=${COMPONENT}_${TESTPLAN}_${DISTRIBUTION}_${STARTED}.log
rm -f ${LOG} > /dev/null

# Log function using tee to split the log statement
log()
{
	echo $1 | tee -a ${CWD}/${LOG}
}

echo `date +%Y-%m-%d_%H-%M-%S` >> ${CWD}/${LOG}
log "Starting incremental performance test of NSP component '${COMPONENT}' using testplan '${TESTPLAN}' and distribution '${DISTRIBUTION}'.."

if [[ ! -r ${HOSTS} ]]; then
	check_errs 3 "Hosts configuration file not found: ${HOSTS}"
fi

# Verify slave configuration and revision
MASTER_REVISION=`${CWD}/slave_test.sh`
check_errs $? "Could not determine master revision."
MASTER_REVISION=`echo "${MASTER_REVISION}" | grep "Revision:" | cut -d' ' -f2`
for ((i=1; i<=${MAX_NODES}; i++)); do
	SLAVE_NODE=`grep slave${i}.host ${HOSTS} | cut -d'=' -f2`
	SLAVE_PATH=`grep slave${i}.path ${HOSTS} | cut -d'=' -f2`
	if [[ -z "${SLAVE_NODE}" ]] || [[ -z "${SLAVE_PATH}" ]]; then
		check_errs 1 "Slave ${i} was not configured correct in ${HOSTS}"
	fi
	REVISION=`ssh ${SLAVE_NODE} "if [[ -f ${SLAVE_PATH}/slave_test.sh ]]; then ${SLAVE_PATH}/slave_test.sh; fi"`
	REVISION=`echo "${REVISION}" | grep "Revision:" | cut -d' ' -f2`
	if [[ "" == "${REVISION}" ]]; then
		check_errs 1 "Slave ${i} has incorrect path in ${HOSTS}"
	fi
	if [[ "${MASTER_REVISION}" != "${REVISION}" ]]; then
		check_errs 1 "Slave ${i} has incorrect revision (was ${REVISION}, exptected ${MASTER_REVISION})"
	fi
done

if [[ -f ${CWD}/master.throughput ]]; then
	rm -f ${CWD}/master.throughput
	check_errs $? "Could not remove throughput log file ${CWD}/master.throughput"
fi

if [[ -f ${CWD}/master.errorpercentage ]]; then
	rm -f ${CWD}/master.errorpercentage
	check_errs $? "Could not remove error percentage log file ${CWD}/master.errorpercentage"
fi

for (( i = THREADS; i <= ${MAX_THREADS}; i = i + ${INC_THREADS} )); do
	THREADS=${i}
	log "Running performance test using ${THREADS} threads on `expr ${NODES} + 1` nodes.."
	${CWD}/master.sh -h ${HOSTS} -p ${PORT} -t ${THREADS} -n ${NODES} ${COMPONENT} ${TESTPLAN} ${DISTRIBUTION} | awk '{printf "\t%s\n", $0}' >> ${CWD}/${LOG}
	check_errs $? "Error running ${THREADS} threads on `expr ${NODES} + 1` nodes"

	if [[ `cat master.throughput | wc -l | tr -d ' '` > 1 ]]; then
		INCREMENT=`cat master.throughput | awk 'BEGIN{TP=0}{if (TP!=0){INC=(($1-TP)/TP)*100}TP=$1}END{printf ("%i", INC)}'`
		# The target increment is 60% of the increment in threads or min 5%
		TARGET=`echo ${THREADS} ${INC_THREADS} | awk '{printf ("%i", (($2/($1-$2))*100)*0.6)}'`
		if [[ ${TARGET} -lt 5 ]]; then
			TARGET=5
		fi
		log "Compared to last run the throughput was increased by ${INCREMENT}% with a target of ${TARGET}%"
		if [[ ${INCREMENT} -lt ${TARGET} ]]; then
			log "Reached maximum throughput by increasing threads. Adding nodes now"
			if [[ ${INCREMENT} -lt 0 ]]; then
				# Roll back the last test since it decreased the throughput
				THREADS=`expr ${THREADS} - ${INC_THREADS}`
				sed -i.bak '$ d' master.throughput > /dev/null
				check_errs $? "Could not remove last run from the master throughput file"
				rm master.throughput.bak > /dev/null
				check_errs $? "Could not remove master throughput backup file"
			fi
			break
		else
			if [[ `expr ${THREADS} + ${INC_THREADS}` -gt ${MAX_THREADS} ]]; then
				check_errs 1 "The maximum number of threads was reached before the throughput could be determined - Contact support."
			fi
		fi
	fi
done

for (( i = NODES + ${INC_NODES}; i <= ${MAX_NODES}; i = i + ${INC_NODES} )); do
	NODES=${i}
	log "Running performance test using ${THREADS} threads on `expr ${NODES} + 1` nodes.."
	${CWD}/master.sh -h ${HOSTS} -p ${PORT} -t ${THREADS} -n ${NODES} ${COMPONENT} ${TESTPLAN} ${DISTRIBUTION} | awk '{printf "\t%s\n", $0}' >> ${CWD}/${LOG}
	check_errs $? "Error running ${THREADS} threads on `expr ${NODES} + 1` nodes"

	INCREMENT=`cat master.throughput | awk 'BEGIN{TP=0}{if (TP!=0){INC=(($1-TP)/TP)*100}TP=$1}END{printf ("%i", INC)}'`
	# The target increment is 60% of the increment in nodes or min 5%
	TARGET=`echo ${NODES} ${INC_NODES} | awk '{printf ("%i", (($2/(($1+1)-$2))*100)*0.6)}'`
	if [[ ${TARGET} -lt 5 ]]; then
		TARGET=5
	fi
	log "Compared to last run the throughput was increased by ${INCREMENT}% with a target of ${TARGET}%"
	if [[ ${INCREMENT} -lt ${TARGET} ]]; then
		log "Reached maximum throughput. Stopping test"
		break
	else
		if [[ `expr ${NODES} + ${INC_NODES}` -gt ${MAX_NODES} ]]; then
			check_errs 1 "The maximum number of nodes was reached before the throughput could be determined - Contact support."
		fi
	fi
done

for (( i = 0; i < ${ADD_ITERATIONS}; i = i + 1 )); do
	THREADS=`expr ${THREADS} + ${INC_THREADS}`
	log "Running additional performance test using ${THREADS} threads on `expr ${NODES} + 1` nodes.."
	${CWD}/master.sh -h ${HOSTS} -p ${PORT} -t ${THREADS} -n ${NODES} ${COMPONENT} ${TESTPLAN} ${DISTRIBUTION} | awk '{printf "\t%s\n", $0}' >> ${CWD}/${LOG}
	check_errs $? "Error running ${THREADS} threads on `expr ${NODES} + 1` nodes"

	INCREMENT=`cat master.throughput | awk 'BEGIN{TP=0}{if (TP!=0){INC=(($1-TP)/TP)*100}TP=$1}END{printf ("%i", INC)}'`
	log "Compared to last run the throughput was increased by ${INCREMENT}%"
	if [[ `expr ${THREADS} + ${INC_THREADS}` -gt ${MAX_THREADS} ]]; then
		check_errs 1 "The maximum number of threads was reached before the throughput could be determined - Contact support."
	fi
done

echo `date +%Y-%m-%d_%H-%M-%S` >> ${CWD}/${LOG}

REPORT_NAME="NO_REPORT"
if [[ -f ${CWD}/master.throughput ]]; then
	THROUGHPUT=`cat ${CWD}/master.throughput | sort -n | tail -1 | cut -d' ' -f1`
	REPORT_NAME=`cat ${CWD}/master.throughput | sort -n | tail -1 | cut -d' ' -f2`
	log "The throughput for the NSP Component '${COMPONENT}' was measured to ${THROUGHPUT} requests per second."
	log "The performance report for this throughput is available in ${REPORT_NAME}"
	rm -f ${CWD}/master.throughput
	check_errs $? "Could not remove throughput log file ${CWD}/master.throughput"
fi

if [[ -f ${CWD}/master.errorpercentage ]]; then
	ERRORPERCENTAGE=`cat master.errorpercentage | grep "${REPORT_NAME}" | cut -d' ' -f1`
	if [[ -n "${ERRORPERCENTAGE}" ]]; then
		log "Note: ${ERRORPERCENTAGE}% of the samples failed."
	fi
	rm -f ${CWD}/master.errorpercentage
	check_errs $? "Could not remove error percentage log file ${CWD}/master.errorpercentage"
fi

if [[ "${REPORT_NAME}" != "NO_REPORT" ]]; then
	mv ${CWD}/${LOG} ${REPORT_NAME}.log
	check_errs $? "Could not remove throughput log file ${CWD}/master.throughput"
	echo "The log of this incremental performance test is available in ${REPORT_NAME}.log"
fi

popd > /dev/null
