#!/bin/bash

# Source the profile script which should contain at least a JAVA_HOME
if [ -f ~/.nsp_perf_profile ]; then
	. ~/.nsp_perf_profile
fi

# Ensure JAVA_HOME is set
if [ -z "$JAVA_HOME" ]; then
	echo "Error: JAVA_HOME environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export JAVA_HOME

# Ensure MAVEN_HOME is set
if [ -z "$MAVEN_HOME" ]; then
	echo "Error: MAVEN_HOME environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export MAVEN_HOME

# Ensure M3 is set
if [ -z "$M3" ]; then
	echo "Error: M3 environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export M3

# Ensure SVN_HOME is set
if [ -z "$SVN_HOME" ]; then
	echo "Error: SVN_HOME environment variable is not set. Must be defined in ~/.nsp_perf_profile"
	exit 1
fi
export SVN_HOME

SCRIPT_NAME=`basename $0`

# for checking errors - and exiting if any - call as
#  check_errs $? 'Some error message'
check_errs()
{
  # Function. Parameter 1 is the return code
  # Para. 2 is text to display on failure.
  if [ "$1" -ne "0" ]; then
	echo "ERROR # ${1} : ${2}"
	exit ${1}
  fi
}

# Print usage
usage()
{
  # Function.
	echo "usage: ${SCRIPT_NAME} <Master> <Port> <Slave number> <Component> <Testplan> <Distribution> <Timestamp> <Hosts file>"
}

# Read mandatory arguments using shift and $1
MASTER=$1
shift
PORT=$1
shift
SLAVE_ID=$1
shift
ACTIVE_NODES=$1
shift
COMPONENT=$1
shift
TESTPLAN=$1
shift
DISTRIBUTION=$1
shift
THREADS=$1
shift
STARTED=$1
shift
HOSTS=$1

# Identification variables
SLAVE=`hostname`

# Build file names
TESTPLAN_NAME=${COMPONENT}_${TESTPLAN}_${DISTRIBUTION}_${MASTER}_slave${SLAVE_ID}_${STARTED}.assembled.jmx
REPORT_NAME=${COMPONENT}_${TESTPLAN}_${DISTRIBUTION}_${MASTER}_slave${SLAVE_ID}_${STARTED}

# A hassle to get current working dir
unamestr=`uname -s`
case "${unamestr}" in
	*Linux*)
		CWD=$(dirname $(readlink -f $0))
	;;
	*Darwin*)
		# readlink works different on OS X so we use a python script instead
		REALPATH=$(python -c 'import os,sys;print os.path.realpath(sys.argv[1])' $0)
		CWD=$(dirname $REALPATH)
	;;
	*CYGWIN*)
		# translate to full cygpath
		CWD=$(cygpath -m $(dirname $(readlink -f $(cygpath $0))))
	;;
	*)
		echo "Unable to detect OS from '${unamestr}'"
		exit 1
	;;
esac
pushd ${CWD} > /dev/null

echo "Preparing performance test of NSP component '${COMPONENT}'.."

find ${CWD} -name "*.assembled.jmx" | xargs rm -f
check_errs $? "Unable to prepare test plan directories"

find ${CWD} -name "*.jtl" | xargs rm -f
check_errs $? "Unable to prepare test result directories"

${M3}/mvn clean install > ${CWD}/mvn.clean.install.log
check_errs $? "Unable to prepare NSP JMeter Extensions, is Maven available?"

JMETER_COMMON=`find ${CWD}/modules/jmeter-common -name "jmeter-common*.jar"`

${JAVA_HOME}/bin/java -cp ${JMETER_COMMON} com.arosii.jmeter.utils.nsp.PrepareTestPlan ${CWD}/tests/${COMPONENT}/src/test/jmeter/templates/testplans/${TESTPLAN}.template.jmx ${CWD}/tests/${COMPONENT}/src/test/jmeter/templates/distributions/${DISTRIBUTION}.template.jmx ${CWD}/tests/${COMPONENT}/src/test/jmeter/${TESTPLAN_NAME} ${THREADS} ${ACTIVE_NODES} ${SLAVE_ID} ${HOSTS} > ${CWD}/java.PrepareTestPlan.log 2>&1
check_errs $? "Unable to create test plan with selected distribution"

# Synchronize with master
echo "Synchronizing execution with master"
${JAVA_HOME}/bin/java -cp ${JMETER_COMMON} com.arosii.jmeter.utils.nsp.sync.SyncSlave ${MASTER} ${PORT} 60 > ${CWD}/java.SyncSlave.log 2>&1
check_errs $? "Unable to synchronize execution with master"

pushd ${CWD}/tests/${COMPONENT} > /dev/null

echo "Running performance test of NSP component '${COMPONENT}', this may take several minutes.."
echo "Using testplan '${TESTPLAN}' and distribution '${DISTRIBUTION}'"
echo "Started at `date +%Y-%m-%d_%H-%M-%S`"
${M3}/mvn verify > ${CWD}/mvn.verify.log
check_errs $? "Unable to run testplan"
echo "Finished at `date +%Y-%m-%d_%H-%M-%S`"

popd > /dev/null

echo "Creating performance report.."

mkdir ${CWD}/${REPORT_NAME}
check_errs $? "Unable to create report directory"

mv ${CWD}/mvn.clean.install.log ${CWD}/${REPORT_NAME}/mvn.clean.install.log
check_errs $? "Unable to collect extension log file"

mv ${CWD}/java.PrepareTestPlan.log ${CWD}/${REPORT_NAME}/java.PrepareTestPlan.log
check_errs $? "Unable to collect test plan preparation log file"

mv ${CWD}/java.SyncSlave.log ${CWD}/${REPORT_NAME}/java.SyncSlave.log
check_errs $? "Unable to collect slave synchronization log file"

mv ${CWD}/${HOSTS} ${CWD}/${REPORT_NAME}/hosts.properties
check_errs $? "Unable to collect hosts file"

mv ${CWD}/mvn.verify.log ${CWD}/${REPORT_NAME}/mvn.verify.log
check_errs $? "Unable to collect performance log file"

mv ${CWD}/tests/${COMPONENT}/target/jmeter/logs/${TESTPLAN_NAME}.log ${CWD}/${REPORT_NAME}/
check_errs $? "Unable to collect JMeter log file"

mv ${CWD}/tests/${COMPONENT}/src/test/jmeter/${TESTPLAN_NAME} ${CWD}/${REPORT_NAME}/
check_errs $? "Unable to collect the test plan"

mv ${CWD}/tests/${COMPONENT}/target/jmeter/results/*.jtl ${CWD}/${REPORT_NAME}/
check_errs $? "Unable to collect JMeter result file"

if [[ -f ${CWD}/tests/${COMPONENT}/summary.jtl ]]; then
	mv ${CWD}/tests/${COMPONENT}/summary.jtl ${CWD}/${REPORT_NAME}/summary.jtl
	check_errs $? "Unable to collect the summary report"
fi

if [[ -f ${CWD}/tests/${COMPONENT}/RegistryResponseValidation.log ]]; then
	mv ${CWD}/tests/${COMPONENT}/RegistryResponseValidation.log ${CWD}/${REPORT_NAME}/RegistryResponseValidation.log
	check_errs $? "Unable to collect RegistryResponseValidation.log"
fi

if [[ -f ${CWD}/tests/${COMPONENT}/RepositoryResponseValidation.log ]]; then
	mv ${CWD}/tests/${COMPONENT}/RepositoryResponseValidation.log ${CWD}/${REPORT_NAME}/RepositoryResponseValidation.log
	check_errs $? "Unable to collect RepositoryResponseValidation.log"
fi

ERRORPERCENTAGE=
THROUGHPUT=
if [[ -f ${CWD}/tests/${COMPONENT}/summary.jtl.throughput ]]; then
	ERRORPERCENTAGE=`head -1 ${CWD}/tests/${COMPONENT}/summary.jtl.throughput`
	THROUGHPUT=`tail -1 ${CWD}/tests/${COMPONENT}/summary.jtl.throughput`
	mv ${CWD}/tests/${COMPONENT}/summary.jtl.throughput ${CWD}/${REPORT_NAME}/summary.jtl.throughput
	check_errs $? "Unable to collect the throughput report"
fi

pushd ${CWD}/${REPORT_NAME} > /dev/null

tar -czf ${CWD}/${REPORT_NAME}.tar.gz . > /dev/null
check_errs $? "Unable to create performance report"

popd > /dev/null

rm -rf ${CWD}/${REPORT_NAME} > /dev/null
check_errs $? "Unable to clean report directory ${REPORT_NAME}"

pushd ${CWD}/tests/${COMPONENT} > /dev/null

${M3}/mvn clean > /dev/null
check_errs $? "Unable to clean working directories"

popd > /dev/null

echo "Performance test of NSP component '${COMPONENT}' completed."
echo "The report is available in the file ${CWD}/${REPORT_NAME}.tar.gz"
if [[ -n "${THROUGHPUT}" ]]; then
	echo "Throughput was measured to ${THROUGHPUT} requests per second."
fi
if [[ -n "${ERRORPERCENTAGE}" ]] && [[ "0" != "${ERRORPERCENTAGE}" ]]; then
	echo "Note: ${ERRORPERCENTAGE}% of the samples failed."
fi

# Synchronize with master
echo "Synchronizing with master"
${JAVA_HOME}/bin/java -cp ${JMETER_COMMON} com.arosii.jmeter.utils.nsp.sync.SyncSlave ${MASTER} ${PORT} 60 > ${CWD}/java.SyncSlave.log 2>&1
check_errs $? "Unable to synchronize repport collection with master"

popd > /dev/null
